import numpy as np
from pysaber.modelssr import SourceBlur,DetectorBlur,get_scale,get_FWHM,combine_psfs,convolve_psf
from scipy.signal import fftconvolve
#from scipy.optimize import minimize_scalar
from scipy.optimize import minimize
from copy import deepcopy    
import sys
from skimage.restoration import wiener
#from matplotlib.widgets import RectangleSelector
import matplotlib.pyplot as plt

def regparam_estimate(recon,prior_weights):
    """
    Maximum likelihood estimate of regularization parameter given a initial reconstruction

    Parameters:
        recon (numpy.ndarray): Initial reconstruction (Blurred radiograph is a potential initial reconstruction)
        prior_weights (numpy.ndarray): 3x3 weights for the prior local neighborhood regularization.

    Returns:
        Estimate of regularization parameter  
    """
    (spat_cost,spat_N) = computePriorCost(recon,prior_weights,1.0)
    regparam = 1.0/(1.2*spat_cost/spat_N)
    return(regparam)
    
def computePriorGradient(recon,weights,regparam):
    """
    Compute gradient of prior cost function.

    Parameters:
        recon (numpy.ndarray): Reconstruction (intermediate deblurred radiograph)
        weights (numpy.ndarray): 3x3 weights for the prior local neighborhood regularization.
        regparam (float): Regularization parameter

    Returns:
        numpy.ndarray: Gradient of prior cost with respect to each pixel in the reconstruction
    """
    sz = np.shape(recon)
    grad = np.zeros(sz,dtype=float)
    if (regparam != 0):
        for i in ([-1,0,1]):
            for j in ([-1,0,1]):
                r_xs = 0; r_xe = sz[1]; rn_xs = 0; rn_xe = sz[1]
                r_ys = 0; r_ye = sz[0]; rn_ys = 0; rn_ye = sz[0]
                if (j == -1): r_xs = 1; rn_xe = sz[1]-1;
                if (i == -1): r_ys = 1; rn_ye = sz[0]-1;
                if (j == 1): r_xe = sz[1]-1; rn_xs = 1;
                if (i == 1): r_ye = sz[0]-1; rn_ys = 1;
            
                if (i != 0 or j != 0):
                    Delta = recon[r_ys:r_ye,r_xs:r_xe] - recon[rn_ys:rn_ye,rn_xs:rn_xe]
                    grad[r_ys:r_ye,r_xs:r_xe] += regparam*weights[i+1,j+1]*1.2*np.sign(Delta)*(np.fabs(Delta)**0.2)

    #print ("Avg of recon is ", np.mean(recon))
    return grad

def computePriorLineSearchParams(recon,direc,weights,regparam):
    """
    Vectorized operations for fast computation of line search.

    Parameters:
        recon (numpy.ndarray): Reconstruction (intermediate deblurred radiograph)
        direc (numpy.ndarray): Descent direction
        weights (numpy.ndarray): 3x3 weights for the prior local neighborhood regularization.
        regparam (float): Prior regularization parameter

    Returns:
        numpy.ndarray: Difference between all pairs of a reconstructed pixel and its neighorhood pixel
        numpy.ndarray: Difference between all pairs of a descent direction pixel and its neighorhood pixel
    """
    sz = np.shape(recon)
    off = np.zeros((sz[0],sz[1],3,3),dtype=float)
    slope = np.zeros((sz[0],sz[1],3,3),dtype=float)
    if (regparam != 0):
        for i in ([0,1]):
            if (i == 1): j_sweep = [-1, 0, 1]
            else: j_sweep = [0, 1]
            for j in j_sweep:
                r_xs = 0; r_xe = sz[1]; rn_xs = 0; rn_xe = sz[1]
                r_ys = 0; r_ye = sz[0]; rn_ys = 0; rn_ye = sz[0]
                if (j == -1): r_xs = 1; rn_xe = sz[1]-1;
                if (i == -1): r_ys = 1; rn_ye = sz[0]-1;
                if (j == 1): r_xe = sz[1]-1; rn_xs = 1;
                if (i == 1): r_ye = sz[0]-1; rn_ys = 1;
                
                if (i != 0 or j != 0):
                    off[r_ys:r_ye,r_xs:r_xe,i,j] = (recon[r_ys:r_ye,r_xs:r_xe] - recon[rn_ys:rn_ye,rn_xs:rn_xe])
                    slope[r_ys:r_ye,r_xs:r_xe,i,j] = (direc[r_ys:r_ye,r_xs:r_xe] - direc[rn_ys:rn_ye,rn_xs:rn_xe])
    return (off,slope)

def computePriorCost(recon,weights,regparam):
    """
    Compute prior cost function.

    Parameters:
        recon (numpy.ndarray): Reconstruction (intermediate deblurred radiograph)
        direc (numpy.ndarray): Descent direction
        regparam (float): Prior regularization parameter

    Returns:
        float: Prior cost value
        int: Number of pairs of neighboring pixels
    """
    sz = np.shape(recon)
    cost = 0.0
    N = 0.0
    if (regparam != 0):
        for i in ([0,1]):
            if (i == 1): j_sweep = [-1, 0, 1]
            else: j_sweep = [0, 1]
            for j in j_sweep:
                    r_xs = 0; r_xe = sz[1]; rn_xs = 0; rn_xe = sz[1]
                    r_ys = 0; r_ye = sz[0]; rn_ys = 0; rn_ye = sz[0]
                    if (j == -1): r_xs = 1; rn_xe = sz[1]-1;
                    if (i == -1): r_ys = 1; rn_ye = sz[0]-1;
                    if (j == 1): r_xe = sz[1]-1; rn_xs = 1;
                    if (i == 1): r_ye = sz[0]-1; rn_ys = 1;
                    
                    if (i != 0 or j != 0):
                        diff = np.fabs(recon[r_ys:r_ye,r_xs:r_xe] - recon[rn_ys:rn_ye,rn_xs:rn_xe])
                        N += np.size(diff)
                        cost += regparam*weights[i+1,j+1]*np.sum(diff**1.2)
    return (cost,N)

def LineSearchCost(x,forw_weights,forw_off,forw_slope,prior_weights,prior_regparam,prior_off,prior_slope):
    """
    Cost for line search optimization.

    Parameters:
        x (float): Step size for line search
        forw_weights (numpy.ndarray): Weights for each pixel in the forward model
        forw_off (numpy.ndarray): Offset term within the linear error term of forward cost 
        forw_slope (numpy.ndarray): Slope term within the linear error term of forward cost
        prior_weights (numpy.ndarray): 3x3 weights for the prior local neighborhood regularization.
        prior_regparam (float): Prior regularization parameter
        prior_off (numpy.ndarray): Offset term within the linear error term of prior cost  
        prior_slope (numpy.ndarray): Slope term within the linear error term of prior cost

    Returns:
        float: Total cost which is the sum of forward and prior cost values 
    """
    forw_cost = forw_off + x*forw_slope
    forw_cost = 0.5*np.sum(forw_cost*forw_cost*forw_weights)
    prior_cost = 0
    if (prior_regparam != 0): #Improve speed for unregularized reconstruction
        prior_cost = prior_off + x*prior_slope
        prior_cost = prior_regparam*np.sum(prior_weights*(np.fabs(prior_cost)**1.2))
    return (forw_cost+prior_cost)

def least_squares_deblur(rad,sod,odd,pix_wid,src_pars,det_pars,reg_par,init_rad=None,weights=None,thresh=1e-4):
    """
    Function to reduce blur (deblur) in radiographs using a regularized least squares iterative algorithm.     

    Parameters:
        rad (numpy.ndarray): Normalized radiograph to deblur
        sod (float): Source to object distance (SOD) of the radiograph
        odd (float): Object to detector distance (ODD) of the radiograph
        pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
        src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
        det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
        reg_par (float): Regularization parameter for the least squares deblurring algorithm. Noise and ringing artifacts in the deblurred radiograph decreases with increasing values for ``reg_par`` and vice versa. But, note that increasing ``reg_par`` can also result in excessive blurring due to over-regulization. It is recommended to empirically choose this parameter by increasing or decreasing it by a factor, greater than one (such as ``2`` or ``10``), until the desired image quality is achieved.
        init_rad (numpy.ndarray): Initial estimate for the deblurred radiograph. If set to ``None``, then the blurred radiograph is used as an initial estimate. If not ``None``, then :attr:`init_rad` must be a numpy.ndarray of the same shape as the radiograph :attr:`rad`. 
        weights (numpy.ndarray): Array of weights of the same shape as :attr:`rad` that is useful to model measurement noise. It can be used to increase or decrease the influence of a particular pixel of :attr:`rad` in the forward model cost function. If set to ``None``, every pixel is assigned the same weight of ``1`` in the cost function.
        thresh (float): Convergence threshold for the minimizer used to deblur the input radiograph. The iterations stop when the ratio of the reduction in the error function (cost value) and the magnitude of the error function is lower than :attr:`thresh`. This is the parameter ``ftol`` that is specified in the ``options`` parameter of ``scipy.optimize.minimize``. The optimizer used is L-BFGS-B.
    
    Returns:
        numpy.ndarray: Deblurred radiograph using regularized least squares algorithm. 
    """
    norm_rad = rad
    max_wid = pix_wid*min(norm_rad.shape)/2
    if weights is None:
        weights = 1.0
    
    src_mod = SourceBlur(pix_wid,max_wid,sod,odd,src_pars['cutoff_FWHM_multiplier'],param_x=get_scale(src_pars['source_FWHM_x_axis'],src_pars['norm_power']),param_y=get_scale(src_pars['source_FWHM_y_axis'],src_pars['norm_power']),param_type='scale',norm_pow=src_pars['norm_power'])
    det_mod = DetectorBlur(pix_wid,max_wid,det_pars['cutoff_FWHM_1_multiplier'],det_pars['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_pars['detector_FWHM_1'],det_pars['norm_power']),param_2=get_scale(det_pars['detector_FWHM_2'],det_pars['norm_power']),weight_1=det_pars['detector_weight_1'],param_type='scale',norm_pow=det_pars['norm_power']) 
        
    blur_psf = combine_psfs(src_mod.get_psf(),det_mod.get_psf(),are_psf=True)
    
    padw = np.array(blur_psf.shape)//2-1
    prior_weights = np.zeros([3,3])
    for i in ([-1,0,1]):
        for j in ([-1,0,1]):
            if (i != 0 or j != 0):
                prior_weights[i+1,j+1] = 1.0/np.sqrt(i**2+j**2)
    prior_weights = prior_weights/np.sum(prior_weights)

    if init_rad is not None:
        deblurred_rad = deepcopy(init_rad)
    else:
        deblurred_rad = deepcopy(norm_rad)
    padw = np.array(blur_psf.shape)//2-1
    rad_shape = norm_rad.shape
    #rad_size = norm_rad.size 
 
    def forw_cost_func(rad):
        pred = convolve_psf(np.pad(rad,padw,'edge'),blur_psf,padw,is_psf=True,warn=False)[padw[0]:-padw[0],padw[1]:-padw[1]]
        #forw_cost = 0.5*np.sum(((norm_rad-pred)**2)*weights)/rad_size
        forw_cost = 0.5*np.sum(((norm_rad-pred)**2)*weights)
        return forw_cost    

    #prior_cost,prior_size = computePriorCost(deblurred_rad,prior_weights,reg_param)
    #print("Initial cost is {}".format(forw_cost_func(deblurred_rad)+prior_cost/prior_size)) 
    #print("Initial cost is {}".format(forw_cost_func(deblurred_rad)+prior_cost)) 
 
    def gradient_func(x):
        x = x.reshape(rad_shape) 
        pred = convolve_psf(np.pad(x,padw,'edge'),blur_psf,padw,is_psf=True,warn=False)[padw[0]:-padw[0],padw[1]:-padw[1]]
        # Initialize error sinogram, e
        pred = (norm_rad-pred)*weights
        grad = -convolve_psf(np.pad(pred,padw,'edge'),blur_psf,padw,is_psf=True,warn=False)[padw[0]:-padw[0],padw[1]:-padw[1]]
        #grad = grad/rad_size+computePriorGradient(x,prior_weights,reg_param)/prior_size
        grad = grad+computePriorGradient(x,prior_weights,reg_par)
        return grad.ravel()

    def cost_func(x): 
        x = x.reshape(rad_shape) 
        prior_cost,_ = computePriorCost(x,prior_weights,reg_par)
        #cost = forw_cost_func(x)+prior_cost/prior_size
        forw_cost = forw_cost_func(x)
        cost = forw_cost+prior_cost
        print("Cost is {}={}+{}".format(cost,forw_cost,prior_cost))
        return cost 

    res = minimize(cost_func,deblurred_rad.ravel(),method='L-BFGS-B',jac=gradient_func,options={'disp':True,'ftol':thresh})
    return res.x.reshape(rad_shape)

def wiener_deblur(rad,sod,odd,pix_wid,src_pars,det_pars,reg_par):
    """
    Function to reduce blur (deblur) in a radiograph using Wiener filtering.     

    Args:
        rad (numpy.ndarray): Normalized radiograph to deblur
        sod (float): Source to object distance (SOD) for the radiograph :attr:`rad`.
        odd (float): Object to detector distance (SDD) for the radiograph :attr:`rad`.
        pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
        src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
        det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
        reg_par (float): Regularization parameter for Wiener filter. Noise and ringing artifacts in the deblurred radiograph decreases with increasing values for ``reg_par`` and vice versa. But, note that increasing ``reg_par`` can also result in excessive blurring due to over-regulization. It is recommended to empirically choose this parameter by increasing or decreasing it by a factor, greater than one (such as ``2`` or ``10``), until the desired image quality is achieved.
    
    Returns:
        numpy.ndarray: Deblurred radiograph using a Wiener filter. 
    """

    max_wid = pix_wid*min(rad.shape)/2
    
    src_mod = SourceBlur(pix_wid,max_wid,sod,odd,src_pars['cutoff_FWHM_multiplier'],param_x=get_scale(src_pars['source_FWHM_x_axis'],src_pars['norm_power']),param_y=get_scale(src_pars['source_FWHM_y_axis'],src_pars['norm_power']),param_type='scale',norm_pow=src_pars['norm_power'])
    det_mod = DetectorBlur(pix_wid,max_wid,det_pars['cutoff_FWHM_1_multiplier'],det_pars['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_pars['detector_FWHM_1'],src_pars['norm_power']),param_2=get_scale(det_pars['detector_FWHM_2'],src_pars['norm_power']),weight_1=det_pars['detector_weight_1'],param_type='scale',norm_pow=det_pars['norm_power']) 
        
    blur_psf = combine_psfs(src_mod.get_psf(),det_mod.get_psf(mix_det=(det_pars['detector_weight_1']!=1)),are_psf=True)
    
    padw = np.array(blur_psf.shape)//2-1
    norm_rad_pad = np.pad(rad,padw,'edge')
    deblur_rad = wiener(norm_rad_pad,blur_psf,float(reg_par),clip=False)
    return deblur_rad.astype(float)[padw[0]:-padw[0],padw[1]:-padw[1]]

"""
xsel1,ysel1,xsel2,ysel2 = None,None,None,None
def callback(click,release):
    global xsel1,ysel1,xsel2,ysel2
    xsel1,ysel1 = click.xdata,click.ydata
    xsel2,ysel2 = release.xdata,release.ydata
    print("Selected region is from ({:.1f},{:.1f}) to ({:.1f},{:.1f})".format(xsel1,ysel1,xsel2,ysel2))

def selector(event):
    print(None)

def least_squares_deblur_old(norm_rad,sod,sdd,pix_wid,src_params,det_params,reg_param,init_rad=None,weights=None,max_iters=100,func_tol=1e-9,compute_cost=True):
    max_wid = pix_wid*min(norm_rad.shape)/2
    
    src_mod = SourceBlur(pix_wid,max_wid,sod,sdd-sod,src_params['cutoff_FWHM_multiplier'],param_x=get_scale(src_params['source_FWHM_x_axis']),param_y=get_scale(src_params['source_FWHM_y_axis']))
    det_mod = DetectorBlur(pix_wid,max_wid,det_params['cutoff_FWHM_1_multiplier'],det_params['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_params['detector_FWHM_1']),param_2=get_scale(det_params['detector_FWHM_2']),weight_1=det_params['detector_weight_1']) 
        
    blur_psf = combine_psfs(src_mod.get_psf(),det_mod.get_psf(),are_psf=True)
    
    padw = np.array(blur_psf.shape)//2-1
    prior_weights = np.zeros([3,3])
    for i in ([-1,0,1]):
        for j in ([-1,0,1]):
            if (i != 0 or j != 0):
                prior_weights[i+1,j+1] = 1.0/np.sqrt(i**2+j**2)
    prior_weights = prior_weights/np.sum(prior_weights)
    prior_regparam = float(reg_param)
     
    #if(prior_regparam == 'ml'):
    #    print("Select a rectangular region with uniform gray values")
    #    fig, ax = plt.subplots()
    #    plt.imshow(blurred_image.values)
    #    selector.RS = RectangleSelector(ax,callback,drawtype='box',useblit=True,button=[1,3],minspanx=5,minspany=5,spancoords='pixels',interactive=True)
    #    plt.show()
    #    variance = np.var(blurred_image.values[int(ysel1):int(ysel2),int(xsel1):int(xsel2)])
    #    prior_param = regparam_estimate(blurred_image.values,prior_weights)
    #    prior_regparam = variance*prior_param
    #    print("Data variance is {} and prior parameter estimate is {}".format(variance,prior_param))
    #    print("Estimated value of regularization parameter is {}".format(prior_regparam))
    #elif not (isinstance(float(prior_regparam),float)):
    #    print("ERROR: Parameter prior_regparam must be a number") 

    #prior_regparam = float(prior_regparam)
    
    cost = np.zeros(max_iters+1)
    forw_cost = np.zeros(max_iters+1)
    prior_cost = np.zeros(max_iters+1)
    bsh = norm_rad.shape
    error = np.zeros(bsh,dtype=float)
    forw_slope = np.zeros(bsh,dtype=float)
    neggrad = np.zeros(bsh,dtype=float)
    update = np.zeros(bsh,dtype=float)
    if weights is None:
        weights = 1.0

    if init_rad is not None:
        deblurred_rad = deepcopy(init_rad)
    else:
        deblurred_rad = deepcopy(norm_rad)
    padw = np.array(blur_psf.shape)//2-1
    
    neggrad_prev = np.zeros(neggrad.shape)
    blurred_pred = convolve_psf(np.pad(deblurred_rad,padw,'edge'),blur_psf,padw,is_psf=True)[padw[0]:-padw[0],padw[1]:-padw[1]]
    # Initialize error sinogram, e
    error = norm_rad-blurred_pred
    if (compute_cost == True):    
        forw_cost[0] = 0.5*np.sum(error*error*weights)
        (prior_cost[0],NS) = computePriorCost(deblurred_rad,prior_weights,prior_regparam)
        cost[0] = forw_cost[0] + prior_cost[0]
        print("Iteration = " + str(0) + ": Cost is " + str(cost[0]) + ", forward cost is " + str(forw_cost[0]) + ", prior cost is " + str(prior_cost[0]))

    for i in range(1,max_iters):    
        blurred_pred = error*weights
        blurred_pred = convolve_psf(np.pad(blurred_pred,padw,'edge'),blur_psf,padw,is_psf=True)[padw[0]:-padw[0],padw[1]:-padw[1]]
        neggrad = blurred_pred
        neggrad -= computePriorGradient(deblurred_rad,prior_weights,prior_regparam)
        if (i <= 1):
            direc = np.copy(neggrad)
        else:
            beta = np.sum(neggrad*(neggrad-neggrad_prev))/np.sum(neggrad_prev*neggrad_prev)
            direc = neggrad + max(0,beta)*direc
        neggrad_prev = np.copy(neggrad)
        #direc = np.copy(neggrad)
        
        blurred_pred = direc
        blurred_pred = convolve_psf(np.pad(blurred_pred,padw,'edge'),blur_psf,padw,is_psf=True)[padw[0]:-padw[0],padw[1]:-padw[1]]
        forw_slope = -blurred_pred
            
        (prior_off,prior_slope) = computePriorLineSearchParams(deblurred_rad,direc,prior_weights,prior_regparam)

        args = (weights,error,forw_slope,prior_weights,prior_regparam,prior_off,prior_slope)
        res = minimize_scalar(LineSearchCost,args=args,method='Brent')
        step = res.x

        #Update reconstruction
        update = step*direc
        deblurred_rad = deblurred_rad + update

        blurred_pred = deblurred_rad
        blurred_pred = convolve_psf(np.pad(blurred_pred,padw,'edge'),blur_psf,padw,is_psf=True)[padw[0]:-padw[0],padw[1]:-padw[1]]
        error = norm_rad-blurred_pred
        
        #Compute cost
        if (compute_cost == True):
            forw_cost[i] = 0.5*np.sum(error*error*weights)
            (prior_cost[i],NS) = computePriorCost(deblurred_rad,prior_weights,prior_regparam)
            cost[i] = forw_cost[i] + prior_cost[i]
                
            print("Iteration = " + str(i) + ": Cost is " + str(cost[i]) + ", forward cost is " + str(forw_cost[i]) + ", prior cost is " + str(prior_cost[i]))
            if (cost[i] > cost[i-1]):
                print("ERROR - Cost increased!")
                print("Increase in cost is " + str(cost[i]-cost[i-1]) + "; Percentage increase is " + str((cost[i]-cost[i-1])/cost[i-1]*100.0))
                sys.exit()
        #percent_update = np.sqrt(np.sum(update*update)/np.sum(recon*recon))*100
        abs_update = np.max(np.fabs(update))
        #print ("Percentage update is ", percent_update)
        #print (step, np.mean(direc), np.mean(forw_grad), np.mean(prior_grad), np.mean(prior_weights))
            
        #print("Percentage change in reconstruction is " + str(percent_update))
        print("Max change in reconstruction is " + str(abs_update))
        if (abs_update < func_tol):
            print("Least squares recon has converged at iteration = " + str(i))
            break
        
    if(i+1 <  max_iters):
        forw_cost[i+1:] = forw_cost[i]
        prior_cost[i+1:] = prior_cost[i]
        cost[i+1:] = cost[i]
            
    return deblurred_rad,cost,forw_cost,prior_cost
"""



import numpy as np
import matplotlib.pyplot as plt
from scipy.fftpack import next_fast_len

def get_FWHM(scale,norm):
    """
    Compute full width half maximum (FWHM) given scale parameter.

    Parameters:
        scale (float): Scale parameter of density function

    Returns:
        float: Full width half maximum (FWHM)
    """
    if scale == 0.0:
        return np.infty
    else:
        return 2.0*(np.log(2)**(1.0/norm))/np.abs(scale)

def get_scale(FWHM,norm):
    """
    Compute scale parameter given full width half maximum (FWHM)

    Parameters:
        FWHM (float): Full width half maximum (FWHM) of density function

    Returns:
        float: Scale parameter
    """
    if FWHM == 0.0:
        return np.infty
    else:
        return 2.0*(np.log(2)**(1.0/norm))/np.abs(FWHM)

class Blur:
    """Class for modeling blur."""
    def __init__(self,delta):
        """
        Constructor for Blur class.

        Parameters:
            delta (float): Pixel width.
        """
        self.delta = delta

    def set_pars(self,cutoff_width,psf_func,psf_grad_func):
        """
        Set functions that will be used to compute point spread function (PSF) and gradient of PSF. 

        Parameters:
            cutoff_width (float): Maximum width beyond which PSF is assumed/clipped to zero
            psf_func (python function): Function to compute PSF
            psf_grad_func (python function): Function to compute gradient of PSF 
        """
        self.cutoff_width = cutoff_width
        self.psf_func = psf_func
        self.psf_grad_func = psf_grad_func

    def get_psf(self,delta=None,cutoff_width=None,psf_func=None,**kwargs):
        """
            Function to compute point spread function (PSF)

            Parameters:
                delta (float): Pixel width.
                cutoff_width (float): Maximum width beyond which PSF is assumed/clipped to zero.
                psf_func (python function): Function to compute PSF
    
            Returns:
                numpy.ndarray: 2D point spread function (PSF) 
        """
        psf_func = self.psf_func if psf_func is None else psf_func
        delta = self.delta if delta is None else delta
        cutoff_width = self.cutoff_width if cutoff_width is None else cutoff_width

        coord = np.arange(0,cutoff_width,delta)
        coord = np.concatenate((-coord[1:][::-1],coord))
        coord_x,coord_y = np.meshgrid(coord,coord,indexing='xy')
        
        if cutoff_width > delta:
            psf = psf_func(coord_x,coord_y,**kwargs)
        else:
            psf = np.array([[0,0,0],[0,1,0],[0,0,0]])

        return psf
    
    def get_grad_psfs(self,delta=None,cutoff_width=None,grad_func=None,**kwargs):
        """
            Function to compute gradient of point spread function (PSF).

            Parameters:
                delta (float): Pixel width.
                cutoff_width (float): Maximum width beyond which PSF is assumed/clipped to zero.
                grad_func (python function): Function to compute gradient of PSF

            Returns:
                list of numpy.ndarray: Gradient of PSF in 2D w.r.t. each PSF parameter.
        """       
 
        grad_func = self.psf_grad_func if grad_func is None else grad_func
        delta = self.delta if delta is None else delta
        cutoff_width = self.cutoff_width if cutoff_width is None else cutoff_width

        coord = np.arange(0,cutoff_width,delta)
        coord = np.concatenate((-coord[1:][::-1],coord))
        coord_x,coord_y = np.meshgrid(coord,coord,indexing='xy')

        psfs = grad_func(coord_x,coord_y,**kwargs)
        if cutoff_width <= delta:
            for i in range(len(psfs)):
                psfs[i] = np.zeros((3,3),dtype=float) 

        return psfs
    
def combine_psfs(psf_1,psf_2,are_psf=False):
    """
        Function to convolve two 2D arrays.

        Parameters:
            psf_1 (numpy.ndarray): First 2D array
            psf_2 (numpy.ndarray): Second 2D array
            are_psf (bool): If true, the 2D array is a PSF and gradient of PSF otherwise.

        Returns:
            numpy.ndarray: Convolution of the input 2D arrays
    """
    
    shape_1 = np.array(psf_1.shape)
    shape_2 = np.array(psf_2.shape)
    conv_shape = shape_1+shape_2-1
  
    fast_shape = np.array([next_fast_len(int(conv_shape[0])),next_fast_len(int(conv_shape[1]))]) 
 
    new_psf = np.fft.irfft2(np.fft.rfft2(psf_1,s=fast_shape)*np.fft.rfft2(psf_2,s=fast_shape),s=fast_shape)
    new_psf = new_psf[:conv_shape[0],:conv_shape[1]]

    if are_psf:
        assert np.isclose(np.sum(new_psf),1.0),"Sum of PSF is {}".format(np.sum(new_psf))
        assert np.isclose(new_psf[new_psf.shape[0]//2,new_psf.shape[1]//2],np.max(new_psf)),"Difference is {}".format(new_psf[new_psf.shape[0]//2,new_psf.shape[1]//2]-np.max(new_psf))
    return new_psf

def convolve_psf(img,psf,pad_width,is_psf=False,pad_type='edge',pad_constant=0,warn=True):
    """
        Function to convolve an image with a point spread function (PSF) or its gradient.

        Parameters:
            img (numpy.ndarray): Input image
            psf (numpy.ndarray): PSF to convolve
            pad_width (float): Maximum width beyond which PSF is assumed/clipped to zero. Padding already applied to input.
            is_psf (float): If true, 2nd parameter is PSF and gradient of PSF otherwise
            pad_type (str): Type of padding used by the function numpy.pad
            pad_constant (float): Constant value to be used as padding by the function numpy.pad

        Returns:
            numpy.ndarray: Convolution of the first two inputs
    """
    if psf.shape[0]>img.shape[0] or psf.shape[1]>img.shape[1]:
        print(psf.shape,img.shape)

    assert psf.shape[0]<=img.shape[0]
    assert psf.shape[1]<=img.shape[1]
    if is_psf:
        assert np.isclose(psf[psf.shape[0]//2,psf.shape[1]//2],np.max(psf))

    padw = np.array(psf.shape)//2
    paddiff = (max(0,padw[0]-pad_width[0]),max(0,padw[1]-pad_width[1]))
    if(paddiff[0] > 0 and warn):
        print("WARNING: y-axis padding is {}. Expected padding of at least {}. Consider increasing the amount of padding. Will use additional {} padding.".format(pad_width[0],padw[0],pad_type))
    if(paddiff[1] > 0 and warn):
        print("WARNING: x-axis padding is {}. Expected padding of at least {}. Consider increasing the amount of padding. Will use additional {} padding.".format(pad_width[1],padw[1],pad_type))
        
    if pad_type is not None: 
        if paddiff[0]>0 or paddiff[1]>0:
            if pad_type == 'constant':
                img = np.pad(img,((paddiff[0],paddiff[0]),(paddiff[1],paddiff[1])),mode=pad_type,constant_values=pad_constant)
            else:
                img = np.pad(img,((paddiff[0],paddiff[0]),(paddiff[1],paddiff[1])),mode=pad_type)
    
    shape_1 = np.array(img.shape)
    shape_2 = np.array(psf.shape)
    conv_shape = shape_1+shape_2-1
    fast_shape = np.array([next_fast_len(int(conv_shape[0])),next_fast_len(int(conv_shape[1]))]) 

    blurimg = np.fft.irfft2(np.fft.rfft2(img,s=fast_shape)*np.fft.rfft2(psf,s=fast_shape),s=fast_shape)
    blurimg = blurimg[:conv_shape[0],:conv_shape[1]]
    blurimg = blurimg[shape_2[0]//2:-(shape_2[0]//2),shape_2[1]//2:-(shape_2[1]//2)]    
   
    assert np.all(np.array(blurimg.shape)==shape_1)
 
    if pad_type is not None:
        if(paddiff[0] > 0):
            blurimg = blurimg[paddiff[0]:-paddiff[0]]
        if(paddiff[1] > 0):
            blurimg = blurimg[:,paddiff[1]:-paddiff[1]]

    return blurimg 
 
class SourceBlur(Blur):
    """Class for modeling X-ray source blur."""
    def __init__(self,delta,max_width,sod,odd,cutoff_FWHM,param_x,param_y,param_type,norm_pow,warn=True):
        """
            Constructor for creating an object of SourceBlur class.

            Parameters:
                delta (float): Pixel width.
                max_width (float): Maximum width beyond which point spread function (PSF) is assumed/clipped to zero.
                sod (float): X-ray source to object distance
                odd (float): Object to detector distance
                cutoff_FWHM (float): Multiple of FWHM beyond which PSF is assumed/clipped to zero.
                param_x (float): Scale/FWHM parameter along x-axis
                param_y (float): Scale/FWHM parameter along y-axis
                param_type (str): If 'scale', then param_x/param_y are scale parameters and FWHM parameters otherwise 
        """
        self.sod = sod
        self.odd = odd
        self.warn = warn
        self.max_width = max_width
        self.cutoff_FWHM = cutoff_FWHM
        self.norm_pow = norm_pow
        super().__init__(delta)
        self.set_params(param_x,param_y,param_type)
        
    def psf_function(self,sod=None,odd=None,scale_x=None,scale_y=None):
        """
            Creates a function to compute point spread function (PSF)

            Parameters:
                sod (float): X-ray source to object distance
                odd (float): Object to detector distance
                scale_x (float): Scale parameter along x-axis
                scale_y (float): Scale parameter along y-axis

            Returns:
                python function: Function to compute PSF             
        """
        scale_x = self.scale_x if scale_x is None else scale_x
        scale_y = self.scale_y if scale_y is None else scale_y
        sod = self.sod if sod is None else sod
        odd = self.odd if odd is None else odd
        def psf_func(x,y):
            num = np.exp(-((x*scale_x*sod/odd)**2+(y*scale_y*sod/odd)**2)**(self.norm_pow/2.0))
            return num/np.sum(num)
        return psf_func

    def psf_grad_function(self,sod=None,odd=None,scale_x=None,scale_y=None):
        """
            Creates a function to compute gradient of the point spread function (PSF)

            Parameters:
                sod (float): X-ray source to object distance
                odd (float): Object to detector distance
                scale_x (float): Scale parameter along x-axis
                scale_y (float): Scale parameter along y-axis

            Returns:
                python function: Function to compute gradient of PSF             
        """
        scale_x = self.scale_x if scale_x is None else scale_x
        scale_y = self.scale_y if scale_y is None else scale_y
        sod = self.sod if sod is None else sod
        odd = self.odd if odd is None else odd

        dist_func = lambda x,y: ((x*scale_x)**2+(y*scale_y)**2)
        def grad_func(x,y):
            sz = x.shape
            assert x[sz[0]//2,sz[1]//2] == 0
            assert y[sz[0]//2,sz[1]//2] == 0
            dist = dist_func(x,y)
            dist_grad = dist.copy()
            dist_grad[sz[0]//2,sz[1]//2] = 1.0 #To prevent divide by zero.
            dist_grad = dist_grad**(self.norm_pow/2.0-1)
            #dist_grad[sz[0]//2,sz[1]//2] = 1.0
            exp = np.exp(-((sod/odd)**(self.norm_pow))*(dist**(self.norm_pow/2.0)))
            exp_sum = np.sum(exp)
            
            const = self.norm_pow*((sod/odd)**self.norm_pow)
            grad_x = -const*exp*scale_x*(x**2)
            grad_x = grad_x*dist_grad
            assert grad_x[sz[0]//2,sz[1]//2] == 0.0
            grad_quo_x = (exp_sum*grad_x-exp*np.sum(grad_x))/(exp_sum**2)

            grad_y = -const*exp*scale_y*(y**2)
            grad_y = grad_y*dist_grad
            assert grad_y[sz[0]//2,sz[1]//2] == 0.0
            grad_quo_y = (exp_sum*grad_y-exp*np.sum(grad_y))/(exp_sum**2)
        
            assert np.isclose(grad_quo_x[sz[0]//2,sz[1]//2],-np.sum(grad_x)/(exp_sum**2)),(grad_quo_x[sz[0]//2,sz[1]//2]+np.sum(grad_x)/(exp_sum**2)) 
            assert np.isclose(grad_quo_y[sz[0]//2,sz[1]//2],-np.sum(grad_y)/(exp_sum**2)),(grad_quo_y[sz[0]//2,sz[1]//2]+np.sum(grad_y)/(exp_sum**2)) 

            return [grad_quo_x,grad_quo_y] 
        return grad_func
 
    def set_params(self,param_x,param_y,param_type='scale'):
        """
            Function to set parameters.

            Parameters:
                param_x (float): Scale/FWHM parameter along x-axis
                param_y (float): Scale/FWHM parameter along y-axis
                param_type (str): If 'scale', then param_x/param_y are scale parameters and FWHM parameters otherwise 
        """
        param_x = np.abs(param_x)
        param_y = np.abs(param_y)
        if param_type == 'FWHM':
            self.FWHM_x,self.FWHM_y = param_x,param_y
            self.scale_x = get_scale(self.FWHM_x,self.norm_pow)
            self.scale_y = get_scale(self.FWHM_y,self.norm_pow)
        elif param_type == 'scale':
            self.scale_x,self.scale_y = param_x,param_y
            self.FWHM_x = get_FWHM(self.scale_x,self.norm_pow)
            self.FWHM_y = get_FWHM(self.scale_y,self.norm_pow)
        else:
            raise ValueError("param_type is invalid")
    
        cutoff_width = (self.odd/self.sod)*self.cutoff_FWHM*max([self.FWHM_x,self.FWHM_y])/2.0
        #print("src blur: cutoff_width is {}".format(cutoff_width))
        if cutoff_width>self.max_width and self.warn:
            print("WARN: The specified maximum width {} for source PSF is less than the cutoff width {} at SOD {} and ODD {}".format(self.max_width,cutoff_width,self.sod,self.odd))
        cutoff_width = min(cutoff_width,self.max_width)
        #cutoff_width = self.max_width

        psf_func = self.psf_function() 
        psf_grad_funcs = self.psf_grad_function() 
        
        super().set_pars(cutoff_width,psf_func,psf_grad_funcs)

    def get_psf_at_source(self):
        """
            Function to compute PSF at the plane of X-ray source.

            Parameters: 
                None

            Returns:
                numpy.ndarray: 2D PSF at plane of X-ray source
        """
        psf_func = self.psf_function(sod=1.0,odd=1.0)
        cutoff_width = self.cutoff_FWHM*max([self.FWHM_x,self.FWHM_y])/2.0
        psf = self.get_psf(self.delta,cutoff_width,psf_func=psf_func)
        return psf
 
class DetectorBlur(Blur):
    """Class for modeling detector blur."""
    def __init__(self,delta,max_width,cutoff_FWHM_1,cutoff_FWHM_2,param_1,param_2,weight_1,param_type,norm_pow,warn=True):
        """
            Constructor for creating an object of DetectorBlur class.

            Parameters:
                delta (float): Pixel width.
                max_width (float): Maximum width beyond which point spread function (PSF) is assumed/clipped to zero.
                cutoff_FWHM_1 (float): Multiple of FWHM_1 for determining cutoff width.
                cutoff_FWHM_2 (float): Multiple of FWHM_2 for determining cutoff width.
                param_1 (float): Scale/FWHM parameter of first exponential
                param_2 (float): Scale/FWHM parameter of second exponential
                weight_1 (float): Weight for first exponential. 
                param_type (str): If 'scale', then param_x/param_y are scale parameters and FWHM parameters otherwise 
        """
        super().__init__(delta)
        self.max_width = max_width
        self.cutoff_FWHM_1 = cutoff_FWHM_1
        self.cutoff_FWHM_2 = cutoff_FWHM_2
        self.norm_pow = norm_pow
        self.warn = warn
        self.set_params(param_1,param_2,weight_1,param_type)

    def psf_function(self,scale_1=None,scale_2=None,weight_1=None):
        """
            Creates a function to compute point spread function (PSF)

            Parameters:
                scale_1 (float): Scale parameter of first exponential
                scale_2 (float): Scale parameter of second exponential
                weight_1 (float): Weight parameter for first exponential

            Returns:
                python function: Function to compute PSF             
        """
        scale_1 = self.scale_1 if scale_1 is None else scale_1
        scale_2 = self.scale_2 if scale_2 is None else scale_2
        p = self.weight_1 if weight_1 is None else weight_1
        def psf_func(x,y,mix_det=True):
            exp_1 = np.exp(-((scale_1*x)**2+(scale_1*y)**2)**(self.norm_pow/2.0))
            if mix_det:
                exp_2 = np.exp(-((scale_2*x)**2+(scale_2*y)**2)**(self.norm_pow/2.0))
                psf_eff = p*exp_1/np.sum(exp_1)+(1.0-p)*exp_2/np.sum(exp_2)
            else:
                psf_eff = exp_1/np.sum(exp_1)
            #print("det blur: PSF max is {}".format(np.max(psf_eff)))
            return psf_eff
        return psf_func
    
    def psf_grad_function(self,scale_1=None,scale_2=None,weight_1=None):
        """
            Creates a function to compute gradient of the point spread function (PSF)

            Parameters:
                scale_1 (float): Scale parameter of first exponential
                scale_2 (float): Scale parameter of second exponential
                weight_1 (float): Weight parameter for first exponential

            Returns:
                python function: Function to compute gradient of PSF             
        """
        scale_1 = self.scale_1 if scale_1 is None else scale_1
        scale_2 = self.scale_2 if scale_2 is None else scale_2
        p = self.weight_1 if weight_1 is None else weight_1
        
        dist_func = lambda x,y: (x**2+y**2)**(self.norm_pow/2.0)
        def grad_func(x,y,mix_det=True):
            sz = x.shape
           
            dist = dist_func(x,y)
            exp_1 = np.exp(-(scale_1**self.norm_pow)*dist)
            exp_1_sum = np.sum(exp_1)
            grad_1 = -self.norm_pow*exp_1*dist
            grad_1 = grad_1*(scale_1**(self.norm_pow-1)) if self.norm_pow!=1 else grad_1
            assert grad_1[sz[0]//2,sz[1]//2]==0.0
            grad_1 = p*(exp_1_sum*grad_1-exp_1*np.sum(grad_1))/(exp_1_sum**2)

            if mix_det:
                exp_2 = np.exp(-(scale_2**self.norm_pow)*dist)
                exp_2_sum = np.sum(exp_2)
                grad_2 = -self.norm_pow*exp_2*dist
                grad_2 = grad_2*(scale_2**(self.norm_pow-1)) if self.norm_pow!=1 else grad_2
                assert grad_2[sz[0]//2,sz[1]//2]==0.0
                grad_2 = (1-p)*(exp_2_sum*grad_2-exp_2*np.sum(grad_2))/(exp_2_sum**2)

                grad_p = exp_1/exp_1_sum - exp_2/exp_2_sum            
                return [grad_1,grad_2,grad_p]
            else:
                return [grad_1]

        return grad_func 
 
    def set_params(self,param_1,param_2,weight_1,param_type='scale'):
        """
            Function to set parameters.

            Parameters:
                param_1 (float): Scale parameter of first exponential
                param_2 (float): Scale parameter of second exponential
                weight_1 (float): Weight parameter for first exponential
                param_type (str): If 'scale', then param_1/param_2 are scale parameters and FWHM parameters otherwise 
        """
        param_1 = np.abs(param_1)
        if param_type == 'FWHM':
            self.FWHM_1 = param_1
            self.scale_1 = get_scale(self.FWHM_1,self.norm_pow)
        elif param_type == 'scale':
            self.scale_1 = param_1
            self.FWHM_1 = get_FWHM(self.scale_1,self.norm_pow)
        else:
            raise ValueError("param_type is invalid")

        if param_2 is not None:
            param_2 = np.abs(param_2)
            if param_type == 'FWHM':
                self.FWHM_2 = param_2
                self.scale_2 = get_scale(self.FWHM_2,self.norm_pow)
            elif param_type == 'scale':
                self.scale_2 = param_2
                self.FWHM_2 = get_FWHM(self.scale_2,self.norm_pow)
            else:
                raise ValueError("param_type is invalid")
            
        if weight_1 is not None:
            weight_1 = 0.0 if weight_1<0.0 else weight_1
            weight_1 = 1.0 if weight_1>1.0 else weight_1
            self.weight_1 = weight_1   
 
        cutoff_width = max(self.cutoff_FWHM_1*self.FWHM_1,self.cutoff_FWHM_2*self.FWHM_2)/2.0
        if cutoff_width>self.max_width and self.warn:
            print("WARN: The maximum width {} specified for detector PSF is less than the cutoff width {}".format(self.max_width,cutoff_width))
        cutoff_width = min(cutoff_width,self.max_width)

        psf_func = self.psf_function() 
        psf_grad_funcs = self.psf_grad_function() 
        
        super().set_pars(cutoff_width,psf_func,psf_grad_funcs)
        
class Transmission:
    """Class for modeling the X-ray transmission function"""
    def __init__(self,trans_model,params):
        """
            Constructor for initializing an object of Transmission class

            Parameters:
                trans_model (list): Transmission model for the sample
                params (list): Parameters of the transmission model  
        """
        self.norm_rad_mask = trans_model['norm_rad_mask']
        self.ideal_trans = trans_model['ideal_trans']
        self.ideal_trans_mask = trans_model['ideal_trans_mask']
        self.ideal_trans_grad = trans_model['ideal_trans_grad']
        self.len_params = len(params)
        self.psf_max_halfsize = np.min(np.array(self.ideal_trans_mask.shape)-np.array(self.norm_rad_mask.shape))//4 #Max half size of psf
        #Divide by 4 because one divide by 2 since there are two PSFs (source and detector blur) and another divide by 2 to get one-sided max width
        #The impulse response from pixels on one side should not leak into the region of interest (unmasked and optimized) pixels on the other side (reflected side) 
        self.set_params(params)

    def set_params(self,params):
        """
            Set parameters of transmission model.

            Parameters:
                params (list): Parameters of the transmission model
        """
        self.params = params.copy()

    def get_trans(self):
        """
            Returns ideal transmission function and masks for normalized radiograph and ideal transmission function

            Parameters:
                None

            Returns:
                numpy.ndarray,numpy.ndarray,numpy.ndarray: Mask for normalized radiograph, ideal transmission function, mask for ideal transmission function
        """
        return self.norm_rad_mask,self.ideal_trans(self.params),self.ideal_trans_mask

    def get_grad(self):
        """
            Returns gradient of transmission function w.r.t. transmission function parameters

            Parameters:
                None

            Returns:
                numpy.ndarray: Gradient of ideal transmission function
        """
        return self.ideal_trans_grad(self.params) 



__docformat__ = 'reStructuredText'

import time
import numpy as np
import yaml
import os
from pysaber.transsr import ideal_trans_sharp_edge,ideal_trans_perp_corner
from pysaber.modelssr import SourceBlur,DetectorBlur,Transmission,get_scale,get_FWHM,combine_psfs,convolve_psf
from pysaber.optimsr import error_function,jacobian_function,set_model_params
from scipy.optimize import minimize,check_grad,approx_fprime
import matplotlib.pyplot as plt

def estimate_blur(rads,sod,odd,pix_wid,edge,thresh=1e-6,pad=[3,3],masks=None,bdary_mask=5.0,perp_mask=5.0,power=1.0,save_dir='./',only_src=False,only_det=False,mix_det=True):
    """
    Estimate parameters of point spread functions (PSF) that model X-ray source blur and/or detector blur from normalized radiographs of a straight sharp edge or mutually perpendicular intersecting pair of sharp edges. 

    This function is used to estimate parameters of the PSFs that model X-ray source blur and/or detector blur. It takes as input the normalized radiographs at multiple source to object distances (SOD) and object to detector distances (ODD). If each radiograph has a single straight edge, then the measurement must be repeated for two different, preferably perpendicular, orientations of the edge. If the radiograph consists of two intersecting perpendicular edges, then a single radiograph at each specified SOD/ODD is sufficient. Simultaneous estimation of source and detector blur will require radiographs at a minimum of two different value pairs for SOD/ODD. During PSF parameter estimation, the influence of certain regions within each radiograph can be removed by masking. For more details, please read ahead and also refer to the documents listed in :ref:`sec_refs`.  

    Args:
        rads (list): List of radiographs, each of type numpy.ndarray, at various SODs and ODDs. Each radiograph must be normalized using the bright-field (also called flat-field) and dark-field images.
        sod (list): List of source to object distances (SOD), each of type *float*, at which each corresponding radiograph in the list :attr:`rads` was acquired. 
        odd (list): List of object to detector distances (ODD), each of type *float*, at which each corresponding radiograph in the list :attr:`rads` was acquired.
        pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
        edge (str): Used to indicate whether there is a single straight edge or two mutually perpendicular edges in each radiograph. If :attr:`edge` is ``perpendicular``, then each radiograph is assumed to have two mutually perpendicular edges. If it is ``straight``, then each radiograph is assumed to have a single straight edge. Only ``perpendicular`` and ``straight`` are legal choices for :attr:`edge`. 
        thresh (float): Convergence threshold for the minimizer during parameter estimation. The iterations stop when the ratio of the reduction in the error function (cost value) and the magnitude of the error function is lower than :attr:`thresh`. This is the parameter :attr:`ftol` that is specified in the :attr:`options` parameter of ``scipy.optimize.minimize``. The optimizer used is L-BFGS-B. During joint estimation of source and detector blur, the convergence threshold for the minimizer during the first two initialization steps is ten times this value. 
        pad (list): List of two integers that determine the amount of padding that must be applied to the radiographs to reduce aliasing during convolution. The number of rows/columns after padding is equal to :attr:`pad_factor[0]`/:attr:`pad_factor[1]` times the number of rows/columns in each radiograph before padding. For example, if the first element in :attr:`pad_factor` is ``2``, then the radiograph is padded to twice its size along the first dimension. 
        masks (list): List of boolean masks, each of type numpy.ndarray and same shape as the radiograph, that is used to exclude pixels from blur estimation. This is in addition to the masking specified by :attr:`bdary_mask` and :attr:`perp_mask`. An example use case is if some pixels in the radiograph :attr:`rads[i]` are bad, then those pixels can be excluded from blur estimation by setting the corresponding entries in :attr:`masks[i]` to ``False`` and ``True`` otherwise. If None, no user specified mask is used.
        bdary_mask (float): Percentage of image region in the radiographs as measured from the outer edge going inwards that must be excluded from blur estimation. Pixels are excluded (or masked) beginning from the outermost periphery of the image and working inwards until the specified percentage of pixels is reached.
        perp_mask (float): Percentage of circular region to ignore during blur estimation around the intersecting corner of two perpendicular edges. Ignored if :attr:`edge` is ``straight``.
        power (float): Shape parameter of the density function used to model each PSF. For example, choosing a value of one for :attr:`power` creates an exponential (Laplacian) density function. Choosing a value of two for :attr:`power` creates a Gaussian density function.
        save_dir (str): Directory where estimated parameters are saved in *yaml* file format. Source blur parameters are saved in the file ``source_params.yml`` within the folder :attr:`save_dir`. Similary, detector blur and transmission function parameters are saved as ``detector_params.yml`` and ``transmission_params.yml``.
        only_src (bool): If ``True``, only estimate source blur parameters.
        only_det (bool): If ``True``, only estimate detector blur parameters.
        mix_det (bool): If ``True``, do not use mixture model for detector blur.

    Returns:
        tuple: Tuple of objects containing the estimated parameters. If estimating both source and detector blur parameters, returns the three element tuple (``src_pars``, ``det_pars``, ``tran_pars``). If estimating only source blur parameters, returns the two element tuple (``src_pars``, ``tran_pars``). If estimating only detector blur parameters, returns the two element tuple (``det_pars``, ``tran_pars``). ``src_pars`` and ``det_pars`` are python dictionaries. ``tran_pars`` is a list of lists. 

        ``src_pars`` contains the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 

        ``det_pars`` contains estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. If ``mix_det`` is ``False``, then value for key ``detector_weight_1`` is fixed at ``1`` and value for key ``detector_FWHM_2`` is fixed at ``0``. 

        ``tran_pars`` contains estimated parameters of the transmission function for each input radiograph. This return value is a list of lists, where each inner nested list consists of two parameters of type `float`. These `float` values give the low and high values respectively of the transmission function. The number of nested lists in the returned list equals the number of input radiographs. Note that the transmission function is the normalized radiograph image that would have resulted in the absence of blur and noise.
    """
    start_time = time.time()
    est_src = only_src or not only_det
    est_det = only_det or not only_src
    sdd = [i+j for i,j in zip(sod,odd)]

    trans_models,trans_params,trans_bounds = [],[],[]
    for j in range(len(rads)):
        if edge == 'perpendicular':
            mk = None if masks is None else masks[j]
            trans_dict,params,bounds = ideal_trans_perp_corner(rads[j],bdary_mask_perc=bdary_mask,pad_factor=pad,mask=mk,perp_mask_perc=perp_mask)
        elif edge == 'straight':
            mk = None if masks is None else masks[j]
            trans_dict,params,bounds = ideal_trans_sharp_edge(rads[j],bdary_mask_perc=bdary_mask,pad_factor=pad,mask=mk)
        else:
            raise ValueError('Argument edge must be either perpendicular or straight')        

        #trans_dict has ideal transmission function, masks, etc
        trans_models.append(trans_dict)
        #Add trans_dict to list trans_models 
        trans_params.append(params)
        trans_bounds.append(bounds)

        #plt.imshow(norm_rads[j]+trans_dict['norm_rad_mask'].astype(float))                  
        #plt.show()
 
        #print("Initial estimates of transmission parameters for radiograph {} are {}.".format(j,params))
        #print("{:.2e} mins has elapsed.".format((time.time()-start_time)/60.0))

    #---------------------------- ESTIMATE BLUR MODEL --------------------------------
    #It is recommended to run the blur model estimation in multiple stages
    #First, only estimate the detector blur using radiographs with the largest SODs
    #Radiographs with the largest SODs are expected to have minimal source blur
    sod_avg = sum(sod)/len(sod) #Compute average of all SODs

    #Only include list elements with SOD > average of all SODs
    rads_det,trans_models_det,trans_params_det,trans_bounds_det,sod_det,sdd_det = [],[],[],[],[],[]
    for i in range(len(rads)):
        if only_det:
            rads_det.append(rads[i])
            trans_models_det.append(trans_models[i])
            trans_params_det.append(trans_params[i])
            trans_bounds_det.append(trans_bounds[i])
            sod_det.append(sod[i])
            sdd_det.append(sdd[i])
        elif sod[i] > sod_avg and not only_src:
            rads_det.append(rads[i])
            trans_models_det.append(trans_models[i])
            trans_params_det.append(trans_params[i])
            trans_bounds_det.append(trans_bounds[i])
            sod_det.append(sod[i])
            sdd_det.append(sdd[i])

    #Estimate detector blur parameters
    if only_det:
        print("----------- Estimation of only detector blur model parameters -----------")
        _,det_params_det,trans_params_det,_,cost_det = get_blur_psfs(rads_det,trans_models_det,sod_det,sdd_det,pix_wid,src_est=False,det_est=True,trans_est=True,src_params=None,det_params=None,trans_params=trans_params_det,trans_bounds=trans_bounds_det,norm_pow=power,convg_thresh=thresh,mix_det=mix_det)
        trans_params = trans_params_det
        print("Estimated detector parameters are {}".format(det_params_det))
    elif not only_src:
        print("----------- Init stage: Estimation of only detector blur model parameters -----------")
        _,det_params_det,_,_,cost_det = get_blur_psfs(rads_det,trans_models_det,sod_det,sdd_det,pix_wid,src_est=False,det_est=True,trans_est=False,src_params=None,det_params=None,trans_params=trans_params_det,trans_bounds=trans_bounds_det,norm_pow=power,convg_thresh=thresh*10,mix_det=mix_det)
        print("Estimated detector parameters are {}".format(det_params_det))
    print("{:.2f} mins has elapsed".format((time.time()-start_time)/60.0))

    #Next, only estimate the source blur using radiographs at the smallest source to object distances
    #Only include list elements with SOD < average of all SODs
    rads_src,trans_models_src,trans_params_src,trans_bounds_src,sod_src,sdd_src = [],[],[],[],[],[]
    for i in range(len(rads)):
        if only_src:
            rads_src.append(rads[i])
            trans_models_src.append(trans_models[i])
            trans_params_src.append(trans_params[i])
            trans_bounds_src.append(trans_bounds[i])
            sod_src.append(sod[i])
            sdd_src.append(sdd[i])
        elif sod[i] < sod_avg and not only_det:
            rads_src.append(rads[i])
            trans_models_src.append(trans_models[i])
            trans_params_src.append(trans_params[i])
            trans_bounds_src.append(trans_bounds[i])
            sod_src.append(sod[i])
            sdd_src.append(sdd[i])

    #Estimate source blur parameters
    if only_src:
        print("----------- Estimation of only source blur model parameters -----------")
        src_params_src,_,trans_params_src,_,cost_src = get_blur_psfs(rads_src,trans_models_src,sod_src,sdd_src,pix_wid,src_est=True,det_est=False,trans_est=True,src_params=None,det_params=None,trans_params=trans_params_src,trans_bounds=trans_bounds_src,norm_pow=power,convg_thresh=thresh,mix_det=mix_det)
        trans_params = trans_params_src
        print("Estimated source parameters are {}".format(src_params_src))
    elif not only_det:
        print("----------- Init stage: Estimation of only source blur model parameters -----------")
        src_params_src,_,_,_,cost_src = get_blur_psfs(rads_src,trans_models_src,sod_src,sdd_src,pix_wid,src_est=True,det_est=False,trans_est=False,src_params=None,det_params=None,trans_params=trans_params_src,trans_bounds=trans_bounds_src,norm_pow=power,convg_thresh=thresh*10,mix_det=mix_det)
        print("Estimated source parameters are {}".format(src_params_src))
    print("{:.2f} mins has elapsed".format((time.time() - start_time)/60.0))

    #Write source blur parameters to file
    if est_src:
        filen = 'source_params.yml' if only_src else 'source_params_init.yml'
        with open(os.path.join(save_dir,filen),'w') as fid:
            yaml.safe_dump(src_params_src,fid,default_flow_style=False)

    #Write detector blur parameters to file
    if est_det:
        filen = 'detector_params.yml' if only_det else 'detector_params_init.yml'
        with open(os.path.join(save_dir,filen),'w') as fid:
            yaml.safe_dump(det_params_det,fid,default_flow_style=False)

    #Write transmission function parameters to file
    trans_dict = {}
    for i in range(len(rads)):
        label = 'radiograph_{}'.format(i)
        trans_dict[label] = {}
        trans_dict[label]['min param'] = trans_params[i][0] 
        trans_dict[label]['max param'] = trans_params[i][1] 
    filen = 'transmission_params.yml' if only_src or only_det else 'transmission_params_init.yml'
    with open(os.path.join(save_dir,filen),'w') as fid:
        yaml.safe_dump(trans_dict,fid,default_flow_style=False)

    if only_src:
        with open(os.path.join(save_dir,'cost.yml'),'w') as fid:
            yaml.safe_dump({'error final':cost_src},fid,default_flow_style=False)
        return src_params_src,trans_params_src
    if only_det:
        with open(os.path.join(save_dir,'cost.yml'),'w') as fid:
            yaml.safe_dump({'error final':cost_det},fid,default_flow_style=False)
        return det_params_det,trans_params_det

    #Finally, refine estimates of both source and detector blur using all radiographs
    print("-------- Final stage: Estimation of source, detector and transmission model parameters --------")
    src_params,det_params,trans_params,trans_bounds,cost = get_blur_psfs(rads,trans_models,sod,sdd,pix_wid,src_est=True,det_est=True,trans_est=True,src_params=src_params_src,det_params=det_params_det,trans_params=trans_params,trans_bounds=trans_bounds,norm_pow=power,convg_thresh=thresh,mix_det=mix_det)
    print("Blur estimation is complete.")
    print("Estimated source parameters are {}, detector parameters are {}, and transmission function parameters are {}".format(src_params,det_params,trans_params))
    print("{:.2f} mins has elapsed".format((time.time()-start_time)/60.0))

    #Write source blur parameters to file
    with open(os.path.join(save_dir,'source_params.yml'),'w') as fid:
        yaml.safe_dump(src_params,fid,default_flow_style=False)

    #Write detector blur parameters to file
    with open(os.path.join(save_dir,'detector_params.yml'),'w') as fid:
        yaml.safe_dump(det_params,fid,default_flow_style=False)

    #Write transmission function parameters to file
    trans_dict = {}
    for i in range(len(rads)):
        label = 'radiograph_{}'.format(i)
        trans_dict[label] = {}
        trans_dict[label]['min param'] = trans_params[i][0] 
        trans_dict[label]['max param'] = trans_params[i][1] 
    with open(os.path.join(save_dir,'transmission_params.yml'),'w') as fid:
        yaml.safe_dump(trans_dict,fid,default_flow_style=False)

    with open(os.path.join(save_dir,'cost.yml'),'w') as fid:
        yaml.safe_dump({'error final':cost,'error source init':cost_src,'error detector init':cost_det},fid,default_flow_style=False)

    return src_params,det_params,trans_params

def get_blur_psfs(norm_rads,trans_models,sod,sdd,pix_wid,src_est,det_est,trans_est,src_params,det_params,trans_params,trans_bounds,norm_pow,convg_thresh,mix_det):
    max_mag = max([(sdd_smpl-sod_smpl)/sod_smpl for sdd_smpl,sod_smpl in zip(sdd,sod)])
    min_mag = min([(sdd_smpl-sod_smpl)/sod_smpl for sdd_smpl,sod_smpl in zip(sdd,sod)])
    if src_params is None:
        src_params = {}
        src_params['source_FWHM_x_axis'] = pix_wid/max_mag if src_est else 0.0
        src_params['source_FWHM_y_axis'] = 2*pix_wid/max_mag if src_est else 0.0
        src_params['norm_power'] = norm_pow
        src_params['cutoff_FWHM_multiplier'] = 10

    if det_params is None:
        det_params = {}
        det_params['detector_FWHM_1'] = pix_wid if det_est else 0.0
        if mix_det:
            det_params['detector_FWHM_2'] = 10*pix_wid if det_est else 0.0
            det_params['detector_weight_1'] = 0.9 if det_est else 0.0
        else:
            det_params['detector_FWHM_2'] = 0.0
            det_params['detector_weight_1'] = 1.0
        det_params['norm_power'] = norm_pow
        det_params['cutoff_FWHM_1_multiplier'] = 10
        det_params['cutoff_FWHM_2_multiplier'] = 10
 
    if trans_bounds is None:
        trans_bounds = []
        if trans_params is None:
            print("Assuming a sharp edge radiograph for bounding transmission parameters")
            for i in range(len(norm_rads)):
                trans_bounds.append([[-1.0,0.5],[0.5,2.0]]) 
        else:
            for param in trans_params:
                trans_bounds.append([[None,None] for i in range(len(param))]) 
 
    if trans_params is None:
        print("Assuming a sharp edge radiograph for transmission parameter initialization")
        trans_params = [[0.0,1.0] for i in range(len(norm_rads))] 

    src_mods,trans_mods = [],[]
    for i in range(len(norm_rads)):
        trans_mods.append(Transmission(trans_models[i],trans_params[i]))
        #print(trans_mods[-1].psf_max_halfsize,pix_wid)
        if i>0:
            assert max_wid == pix_wid*trans_mods[-1].psf_max_halfsize
        max_wid = pix_wid*trans_mods[-1].psf_max_halfsize
        src_mods.append(SourceBlur(pix_wid,max_wid,sod[i],sdd[i]-sod[i],src_params['cutoff_FWHM_multiplier'],param_x=get_scale(src_params['source_FWHM_x_axis'],norm_pow),param_y=get_scale(src_params['source_FWHM_y_axis'],norm_pow),param_type='scale',norm_pow=norm_pow,warn=False))
    det_mod = DetectorBlur(pix_wid,max_wid,det_params['cutoff_FWHM_1_multiplier'],det_params['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_params['detector_FWHM_1'],norm_pow),param_2=get_scale(det_params['detector_FWHM_2'],norm_pow),weight_1=det_params['detector_weight_1'],param_type='scale',norm_pow=norm_pow,warn=False) 
 
    args_var = []
    args_bounds = []
    if src_est == True:
        args_var.append(get_scale(src_params['source_FWHM_x_axis'],norm_pow))
        args_bounds.append([1e-10,None])
        args_var.append(get_scale(src_params['source_FWHM_y_axis'],norm_pow))
        args_bounds.append([1e-10,None])
        print("Will optimize source blur parameters")
        
    if det_est == True:
        args_var.append(get_scale(det_params['detector_FWHM_1'],norm_pow))
        args_bounds.append([1e-10,None])
        if mix_det:
            args_var.append(get_scale(det_params['detector_FWHM_2'],norm_pow))
            args_bounds.append([1e-10,None])
            args_var.append(det_params['detector_weight_1'])
            args_bounds.append([0.8,1.0])
        print("Will optimize detector blur parameters")

    if trans_est == True:
        for par in trans_params:
            args_var.extend(par)
        for bound in trans_bounds:
            args_bounds.extend(bound) 
        print("Will optimize transmission function parameters")

#    eps = np.sqrt(np.finfo(float).eps)
#    print("----------------------------")
#    grad_true = jacobian_function(np.array(args_var),src_mods,det_mod,trans_mods,src_est,det_est,trans_est,norm_rads,pix_wid,mix_det)
#    grad_num = approx_fprime(np.array(args_var),error_function,eps,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,norm_rads,pix_wid,mix_det)
#    print("eps is {}".format(eps))
#    print("Error in jacobian at {} is {}. Numerical jacobian is {}. True jacobian is {}".format(args_var,np.sqrt(np.sum((grad_num-grad_true)**2)),grad_num,grad_true))
#    print("###########################")

    cost = None
    if(src_est or det_est or trans_est):
        res = minimize(error_function,args_var,args=(src_mods,det_mod,trans_mods,src_est,det_est,trans_est,norm_rads,pix_wid,mix_det),method='L-BFGS-B',jac=jacobian_function,bounds=args_bounds,options={'disp':True,'ftol':convg_thresh})
        args_opt = res.x.tolist()
        cost = float(res.fun)
    else:
        args_opt = args_var.copy()

    print("Optimized parameters are {}".format(args_opt))    
    set_model_params(args_opt,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,mix_det) 
    
    args_opt.reverse() 
    if src_est == True:
        src_params['source_FWHM_x_axis'] = float(get_FWHM(args_opt.pop(),norm_pow))
        src_params['source_FWHM_y_axis'] = float(get_FWHM(args_opt.pop(),norm_pow))
        #print("Compare",max_wid,src_params['cutoff_FWHM_multiplier']*max(src_params['source_FWHM_x_axis'],src_params['source_FWHM_y_axis'])*max_mag/2)
        if max_wid < src_params['cutoff_FWHM_multiplier']*max(src_params['source_FWHM_x_axis'],src_params['source_FWHM_y_axis'])*max_mag/2:
            print("WARN: The maximum single sided width {:.2e}, as determined by the amount of padding in radiographs, that is allowed for source PSF may be insufficient. This may or may not be a problem. Consider increasing the pad factor if results are inconsistent. An example of a inconsistency is if FWHM changes significantly when pad factor is changed. Increasing padding will increase software run time.".format(max_wid));
         
    if det_est == True:
        det_params['detector_FWHM_1'] = float(get_FWHM(args_opt.pop(),norm_pow))
        if mix_det:
            det_params['detector_FWHM_2'] = float(get_FWHM(args_opt.pop(),norm_pow))
            det_params['detector_weight_1'] = float(args_opt.pop())
        #print("Compare",max_wid,max(det_params['cutoff_FWHM_1_multiplier']*det_params['detector_FWHM_1'],det_params['cutoff_FWHM_2_multiplier']*det_params['detector_FWHM_2'])/2)
        if max_wid < max(det_params['cutoff_FWHM_1_multiplier']*det_params['detector_FWHM_1'],det_params['cutoff_FWHM_2_multiplier']*det_params['detector_FWHM_2'])/2:
            print("WARN: The maximum single sided width {:.2e}, as determined by the amount of padding in radiographs, that is allowed for detector PSF may be insufficient. This may or may not be a problem. Consider increasing the pad factor if results are inconsistent. An example of a inconsistency is if FWHM changes significantly when pad factor is changed. Increasing padding will increase software run time.".format(max_wid));

    if trans_est == True:
        trans_params = []
        for mod in trans_mods:
            trans_params.append([args_opt.pop() for _ in range(mod.len_params)])

    #src_psf = src_mods[0].get_psf_at_source()
    #det_psf = det_mod.get_psf()
   
    #pred_nrads = []
    #dmod_psf = det_mod.get_psf()
    #for j in range(len(norm_rads)):
    #    rad = norm_rads[j]
    #    radm,tran,tranm = trans_mods[j].get_trans()
    #    smod_psf = src_mods[j].get_psf()
    #    pad_widths = (np.array(tran.shape)-np.array(rad.shape))//2 #assumes symmetric padding
    #    new_psf = combine_psfs(smod_psf,dmod_psf,are_psf=True)
    #    pred = convolve_psf(tran,new_psf,pad_widths,is_psf=True)
    #    pred_nrads.append(pred)
    #return src_params,src_psf,det_params,det_psf,trans_params,trans_bounds,pred_nrads,cost

    return src_params,det_params,trans_params,trans_bounds,cost

def get_source_psf(pix_wid,src_pars,sod=1.0,odd=1.0):
    """
    Function to compute the point spread function (PSF) of X-ray source blur in the plane of the X-ray source or the detector. 

    If source to object distance (SOD) is equal to object to detector distance (ODD), then the PSF on the detector plane is same as that on the plane of the X-ray source. If PSF on detector plane is desired, it is required to specify the SOD and ODD. If PSF on source plane is desired, use the default values for SOD and ODD. 

    Args:
        pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
        src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
        sod (float): Source to object distance (SOD). 
        odd (float): Object to detector distance (ODD).
    
    Returns:
        numpy.ndarray: PSF of X-ray source blur.
    """

    max_wid = 0.5*src_pars['cutoff_FWHM_multiplier']*max(src_pars['source_FWHM_x_axis'],src_pars['source_FWHM_y_axis'])*odd/sod
    src_mod = SourceBlur(pix_wid,max_wid,sod,odd,src_pars['cutoff_FWHM_multiplier'],param_x=get_scale(src_pars['source_FWHM_x_axis'],src_pars['norm_power']),param_y=get_scale(src_pars['source_FWHM_y_axis'],src_pars['norm_power']),param_type='scale',norm_pow=src_pars['norm_power'],warn=False)
    return src_mod.get_psf()
     
def get_detector_psf(pix_wid,det_pars):
    """
        Function to compute point spread function (PSF) of detector blur. 

        Args:
            pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
            det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
        
        Returns:
            numpy.ndarray: PSF of detector
    """
    #if max_wid == None:
    max_wid = 0.5*max(det_pars['cutoff_FWHM_1_multiplier']*det_pars['detector_FWHM_1'],det_pars['cutoff_FWHM_2_multiplier']*det_pars['detector_FWHM_2'])
    det_mod = DetectorBlur(pix_wid,max_wid,det_pars['cutoff_FWHM_1_multiplier'],det_pars['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_pars['detector_FWHM_1'],det_pars['norm_power']),param_2=get_scale(det_pars['detector_FWHM_2'],det_pars['norm_power']),weight_1=det_pars['detector_weight_1'],param_type='scale',norm_pow=det_pars['norm_power'],warn=False)
    return det_mod.get_psf()

def get_effective_psf(pix_wid,src_pars,det_pars,sod=1,odd=1): 
    """
        Function to compute the effective point spread function (PSF), which is the convolution of X-ray source and detector PSFs. 
    
        Args:
            pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
            src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
            det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
            sod (float): Source to object distance (SOD). 
            odd (float): Object to detector distance (ODD).
        
        Returns:
            numpy.ndarray: PSF of effective blur in the plane of detector.
    """
    #if max_wid == None:
    max_wid = 0.5*src_pars['cutoff_FWHM_multiplier']*max(src_pars['source_FWHM_x_axis'],src_pars['source_FWHM_y_axis'])*odd/sod
    src_mod = SourceBlur(pix_wid,max_wid,sod,odd,src_pars['cutoff_FWHM_multiplier'],param_x=get_scale(src_pars['source_FWHM_x_axis']),param_y=get_scale(src_pars['source_FWHM_y_axis']),param_type='scale',norm_pow=src_pars['norm_power'])
    max_wid = 0.5*max(det_pars['cutoff_FWHM_1_multiplier']*det_pars['detector_FWHM_1'],det_pars['cutoff_FWHM_2_multiplier']*det_pars['detector_FWHM_2'])
    det_mod = DetectorBlur(pix_wid,max_wid,det_pars['cutoff_FWHM_1_multiplier'],det_pars['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_pars['detector_FWHM_1']),param_2=get_scale(det_pars['detector_FWHM_2']),weight_1=det_pars['detector_weight_1'],param_type='scale',norm_pow=det_pars['norm_power']) 
    smod_psf = src_mod.get_psf()
    dmod_psf = det_mod.get_psf()
    return combine_psfs(smod_psf,dmod_psf,are_psf=True)

def apply_blur_psfs(rad,sod,odd,pix_wid,src_pars,det_pars,padded_widths=[0,0],pad_type='constant',pad_constant=0):
    """
        Function to blur the input radiograph with point spread functions (PSF) of X-ray source and detector blurs.

        This function blurs the input radiograph with X-ray source blur and detector blur with the specified point spread function (PSF) parameters. This function is useful to observe the effect of source and detector blurs on a simulated radiograph.

        Args:
            rad (numpy.ndarray): Radiograph of type numpy.ndarray that is normalized using the bright-field (also called flat-field) and dark-field images.
            sod (float): Source to object distance (SOD) of radiograph. 
            odd (float): Object to detector distance (ODD) of radiograph.
            pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
            src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
            det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
            padded_widths (list): List of two integers that specifies the amount of padding already applied to input radiograph. The first integer specifies the padding applied along the first dimension of radiograph. The second integer specifies the padding applied along the second dimension of radiograph. Assumes ``padded_widths[k]/2`` amount of padding is applied at both the left and right extremities of dimension ``k``, where ``k`` is ``0`` or ``1``.
            pad_type (str): Type of additional padding that must be used if amount of padding specified in :attr:`padded_widths` is insufficient. Supported values are ``edge`` and ``constant``.  
            pad_constant (float): If :attr:`pad_type` is `constant`, specify the constant value that must be padded.
        
        Returns:
            numpy.ndarray: Radiograph that is blurred using X-ray source and detector blurs.
    """
    input_rad = rad 
    max_wid = pix_wid*(min(padded_widths)//2)
    #Divide by 2 since there are two PSFs (source and detector blur)
    src_mod = SourceBlur(pix_wid,max_wid,sod,odd,src_pars['cutoff_FWHM_multiplier'],param_x=get_scale(src_pars['source_FWHM_x_axis'],src_pars['norm_power']),param_y=get_scale(src_pars['source_FWHM_y_axis'],src_pars['norm_power']),param_type='scale',norm_pow=src_pars['norm_power'])
    det_mod = DetectorBlur(pix_wid,max_wid,det_pars['cutoff_FWHM_1_multiplier'],det_pars['cutoff_FWHM_2_multiplier'],param_1=get_scale(det_pars['detector_FWHM_1'],det_pars['norm_power']),param_2=get_scale(det_pars['detector_FWHM_2'],det_pars['norm_power']),weight_1=det_pars['detector_weight_1'],param_type='scale',norm_pow=det_pars['norm_power']) 
    smod_psf = src_mod.get_psf()
    dmod_psf = det_mod.get_psf(mix_det=(det_pars['detector_weight_1']!=1))
    new_psf = combine_psfs(smod_psf,dmod_psf,are_psf=True)
    blurred_rad = convolve_psf(input_rad,new_psf,padded_widths,is_psf=True,pad_type=pad_type,pad_constant=pad_constant)
    return blurred_rad
    
def get_trans_fit(rad,sod,odd,pix_wid,src_pars,det_pars,tran_pars,edge,pad=[3,3]):
    """
        Function to compute the blur model prediction and ideal transmission function for a radiograph with a single straight edge or two mutually perpendicular edges. 

        For a measured radiograph consisting of a straight sharp edge or two mutually perpendicular edges, get the ideal transmission function and the predicted radiograph from the blur model. Here, the blur model is used to model the impact of blur due to X-ray source and detector.
    
        Parameters:
            rad (numpy.ndarray): Normalized radiograph of a straight sharp edge or two mutually perpendicular edges.
            sod (float): Source to object distance (SOD) for the radiograph :attr:`rad`.
            odd (float): Object to detector distance (SDD) for the radiograph :attr:`rad`.
            pix_wid (float): Effective width of each detector pixel. Note that this is the effective pixel size given by dividing the physical width of each detector pixel by the zoom factor of the optical lens.
            src_pars (dict): Dictionary containing the estimated parameters of X-ray source PSF. It consists of several key-value pairs. The value for key ``source_FWHM_x_axis`` is the full width half maximum (FWHM) of the source PSF along the x-axis (i.e., second numpy.ndarray dimension). The value for key ``source_FWHM_y_axis`` is the FWHM of source PSF along the y-axis (i.e., first numpy.ndarray dimension). All FWHMs are for the source PSF in the plane of the X-ray source (and not the plane of the detector). The value for key ``cutoff_FWHM_multiplier`` decides the non-zero spatial extent of the source PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_x_axis']/2`` and ``src_pars['cutoff_FWHM_multiplier']`` times ``src_pars['source_FWHM_y_axis']/2``. 
            det_pars (dict): Dictionary containing the estimated parameters of detector PSF. It consists of several key-value pairs. The value for key ``detector_FWHM_1`` is the FWHM of the first density function in the mixture density model for detector blur. The first density function is the most dominant part of detector blur. The value for key ``detector_FWHM_2`` is the FWHM of the second density function in the mixture density model. This density function has the largest FWHM and models the long running tails of the detector blur's PSF. The value for key ``detector_weight_1`` is between ``0`` and ``1`` and is a measure of the amount of contribution of the first density function to the detector blur. The values for keys ``cutoff_FWHM_1_multiplier`` and ``cutoff_FWHM_2_multiplier`` decide the non-zero spatial extent of the detector PSF. The PSF is clipped to zero beginning at a distance, as measured from the PSF's origin, equal to the maximum of ``det_pars['cutoff_FWHM_1_multiplier']`` times ``det_pars['detector_FWHM_1']/2`` and ``det_pars['cutoff_FWHM_2_multiplier']`` times ``det_pars['detector_FWHM_2']/2``. 
            tran_pars (list): List containing the estimated parameters of the transmission function for the input radiograph. It consists of two parameters of type `float`. These `float` values give the low and high values respectively of the transmission function. Note that the transmission function is the normalized radiograph image that would have resulted in the absence of blur and noise. If not specified (or specified as None), then the best fitting transmission function parameters are estimated using RANSAC regression.
            edge (str): Used to indicate whether there is a single straight edge or two mutually perpendicular edges in each radiograph. If :attr:`edge` is ``perpendicular``, then each radiograph is assumed to have two mutually perpendicular intersecting edges. If it is ``straight``, then each radiograph is assumed to have a single straight edge. Only ``perpendicular`` and ``straight`` are legal choices for :attr:`edge`.
            pad (list): List of two integers that determine the amount of padding that must be applied to the radiographs to reduce aliasing during convolution. The number of rows/columns after padding is equal to :attr:`pad_factor[0]`/:attr:`pad_factor[1]` times the number of rows/columns in each radiograph before padding. For example, if the first element in :attr:`pad_factor` is ``2``, then the radiograph is padded to twice its size along the first dimension. 

        Returns:
            tuple: Tuple of two arrays of type ``numpy.ndarray``. The first array is blurred radiograph as predicted by the blur model. The second array is transmission function, which is the ideal readiograph in the absence of source and detector blur.  
    """
    if edge == 'perpendicular':
        trans_dict,params,bounds = ideal_trans_perp_corner(rad,pad_factor=pad)
    elif edge == 'straight':
        trans_dict,params,bounds = ideal_trans_sharp_edge(rad,pad_factor=pad)
    else:
        raise ValueError('Argument edge must be either perpendicular or straight')        
        
    trans_mod = Transmission(trans_dict,tran_pars)
    _,trans,_ = trans_mod.get_trans()
    pad_widths = (np.array(trans.shape)-np.array(rad.shape))//2 #assumes symmetric padding
    pred_rad = apply_blur_psfs(trans,sod,odd,pix_wid,src_pars,det_pars,padded_widths=pad_widths)
    pred_rad = pred_rad[pad_widths[0]:-pad_widths[0],pad_widths[1]:-pad_widths[1]]
    trans = trans[pad_widths[0]:-pad_widths[0],pad_widths[1]:-pad_widths[1]] 
    return pred_rad,trans

def get_trans_masks(rad,edge,tran_pars=None,pad=[1,1],mask=None,bdary_mask=5.0,perp_mask=5.0):
    """
        Function to compute transmission function and masks for a radiograph with a single straight edge or two mutually perpendicular edges. 

        For a measured radiograph consisting of a straight sharp edge or two mutually perpendicular edges, get the transmission function, mask for transmission function, and mask for radiograph. 
        
        Parameters:
            rad (numpy.ndarray): Normalized radiograph of a straight sharp edge or two mutually perpendicular edges.
            edge (str): Used to indicate whether there is a single straight edge or two mutually perpendicular edges in each radiograph. If :attr:`edge` is ``perpendicular``, then each radiograph is assumed to have two mutually perpendicular edges. If it is ``straight``, then each radiograph is assumed to have a single straight edge. Only ``perpendicular`` and ``straight`` are legal choices for :attr:`edge`.
            tran_pars (list): List containing the estimated parameters of the transmission function for the input radiograph. It consists of two parameters of type `float`. These `float` values give the low and high values respectively of the transmission function. Note that the transmission function is the normalized radiograph image that would have resulted in the absence of blur and noise. If not specified (or specified as None), then the best fitting transmission function parameters are estimated using RANSAC regression. If specified as ``[0,1]``, this function returns the ideal transmission function.
            pad (list): List of two integers that determine the amount of padding that must be applied to the radiographs to reduce aliasing during convolution. The number of rows/columns after padding is equal to :attr:`pad_factor[0]`/:attr:`pad_factor[1]` times the number of rows/columns in each radiograph before padding. For example, if the first element in :attr:`pad_factor` is ``2``, then the radiograph is padded to twice its size along the first dimension. 
            mask (numpy.ndarray): Boolean mask of the same shape as the radiograph that is used to exclude pixels from blur estimation. This is in addition to the masking specified by :attr:`bdary_mask` and :attr:`perp_mask`. An example use case is if some pixels in the radiograph :attr:`rad` are bad, then those pixels can be excluded from blur estimation by setting the corresponding entries in :attr:`mask` to ``False`` and ``True`` otherwise. If None, no user specified mask is used.
            bdary_mask (float): Percentage of image region in the radiographs as measured from the outer edge going inwards that must be excluded from blur estimation. Pixels are excluded (or masked) beginning from the outermost periphery of the image and working inwards until the specified percentage of pixels is reached.
            perp_mask (float): Percentage of circular region to ignore during blur estimation around the intersecting corner of two perpendicular edges. Ignored if :attr:`edge` is ``straight``.

        Returns:
            tuple: Tuple of three arrays each of type ``numpy.ndarray``. The first array is the transmission function, which is the ideal readiograph in the absence of source and detector blur. The second and third arrays are the masks for the transmission function and radiograph respectively. The mask array indicates what pixels must be included (pixel value of ``True``) or excluded (pixel value of ``False``) during blur estimation.  
    """
 
    if edge == 'perpendicular':
        trans_dict,init_params,bounds = ideal_trans_perp_corner(rad,bdary_mask_perc=bdary_mask,pad_factor=pad,mask=mask,perp_mask_perc=perp_mask)
    elif edge == 'straight':
        trans_dict,init_params,bounds = ideal_trans_sharp_edge(rad,bdary_mask_perc=bdary_mask,pad_factor=pad,mask=mask)
    else:
        raise ValueError('Argument edge must be either perpendicular or straight')        

    if tran_pars is None:
        trans_mod = Transmission(trans_dict,init_params)
    else:
        trans_mod = Transmission(trans_dict,tran_pars)
        
    rmask,trans,tmask = trans_mod.get_trans()
    #assert np.all(rmask==tmask)
    return trans,tmask,rmask 



import numpy as np
from pysaber.modelssr import combine_psfs,convolve_psf
 
def set_model_params(args_var,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,mix_det):
    """
    Set parameters of blur and transmission models.

    Parameters:
        args_var (list): List of parameters that must be updated
        src_mods (list): List of objects of class SourceBlur
        det_mod (list): Object of class DetectorBlur
        trans_mods (list): List of objects of class Transmission
        src_est (bool): If true, source parameter estimation is done
        det_est (bool): If true, detector parameter estimation is done
        trans_est (bool): If true, transmission model parameter estimation is done
    """
    args = args_var.copy()
    args.reverse()
    if src_est:
        scale_axis_x = args.pop()
        scale_axis_y = args.pop()
        for model in src_mods:
            model.set_params(scale_axis_x,scale_axis_y)

    if det_est:
        pars = [args.pop(),args.pop(),args.pop()] if mix_det else [args.pop(),None,None]
        det_mod.set_params(*pars)

    if trans_est:
        for model in trans_mods:
            model.set_params([args.pop() for _ in range(model.len_params)])

def jacobian_function(args_var,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,norm_rads,pix_wid,mix_det):
    """
    Compute the Jacobian (derivates) of cost function

    Parameters:
        args_var (numpy.ndarray): Array of parameters of blur and transmission model
        src_mods (list): List of objects of class SourceBlur
        det_mod (list): Object of class DetectorBlur
        trans_mods (list): List of objects of class Transmission
        src_est (bool): If true, source parameter estimation is done
        det_est (bool): If true, detector parameter estimation is done
        trans_est (bool): If true, transmission model parameter estimation is done
        norm_rads (list): List of normalized radiographs
        pix_wid (float): Width of each pixel

    Returns:
        numpy.ndarray: Array of gradients with respect to parameters being optimized         
    """
    args_var = args_var.tolist()
    #print("Evaluating gradient for parameters {}".format(args_var))
    set_model_params(args_var,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,mix_det)
   
    if src_est: 
        src_grad = [0,0]
    if det_est:
        det_grad = [0,0,0] if mix_det else [0]
    if trans_est:
        trans_grad = []

    det_psf = det_mod.get_psf(mix_det=mix_det)
    if det_est:
        det_gradpsf = det_mod.get_grad_psfs(mix_det=mix_det)

    for j in range(len(norm_rads)):
        rad = norm_rads[j]
        radm,tran,tranm = trans_mods[j].get_trans()
        pad_widths = (np.array(tran.shape)-np.array(rad.shape))//2 #assumes symmetric padding
        
        src_psf = src_mods[j].get_psf()
        src_det_psf = combine_psfs(src_psf,det_psf,are_psf=True)
        src_det_psfpred = convolve_psf(tran,src_det_psf,pad_widths,is_psf=True,warn=False)
        
        if src_est:
            src_gradpsf = src_mods[j].get_grad_psfs()
            for i in range(len(src_gradpsf)):
                src_det_grad = combine_psfs(src_gradpsf[i],det_psf)
                src_det_gradpred = convolve_psf(tran,src_det_grad,pad_widths,warn=False)
                #src_grad[i] += -np.mean((rad[radm]-src_det_psfpred[tranm])*src_det_gradpred[tranm])
                src_grad[i] += -np.sum((rad[radm]-src_det_psfpred[tranm])*src_det_gradpred[tranm])
        if det_est:
            for i in range(len(det_gradpsf)):
                src_det_grad = combine_psfs(src_psf,det_gradpsf[i])
                src_det_gradpred = convolve_psf(tran,src_det_grad,pad_widths,warn=False)
                #det_grad[i] += -np.mean((rad[radm]-src_det_psfpred[tranm])*src_det_gradpred[tranm])
                det_grad[i] += -np.sum((rad[radm]-src_det_psfpred[tranm])*src_det_gradpred[tranm])
        if trans_est:
            tgrad = trans_mods[j].get_grad()
            for i in range(len(tgrad)):
                temp = convolve_psf(tgrad[i],src_det_psf,pad_widths,warn=False)
                #trans_grad.append(-np.mean((rad[radm]-src_det_psfpred[tranm])*temp[tranm]))
                trans_grad.append(-np.sum((rad[radm]-src_det_psfpred[tranm])*temp[tranm]))
    grad = [] 
    if src_est:
        grad.extend(src_grad) 
    if det_est:
        grad.extend(det_grad)
    if trans_est:
        grad.extend(trans_grad)
    grad = np.array(grad) 
    grad = grad/len(norm_rads)

#    print("Gradient is ", grad)
    return grad    

def error_function(args_var,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,norm_rads,pix_wid,mix_det):
    """
    Compute error function (also called cost) that measures the magnitude of the difference between the measured radiograph values and its prediction using the blur model.

    Parameters:
        args_var (numpy.ndarray): Array of parameters of blur and transmission model
        src_mods (list): List of objects of class SourceBlur
        det_mod (list): Object of class DetectorBlur
        trans_mods (list): List of objects of class Transmission
        src_est (bool): If true, source parameter estimation is done
        det_est (bool): If true, detector parameter estimation is done
        trans_est (bool): If true, transmission model parameter estimation is done
        norm_rads (list): List of normalized radiographs
        pix_wid (float): Width of each pixel
        
    Returns:
        float: Error function value (also called cost)  
    """
    args_var = args_var.tolist()
    print("Evaluating fit for parameters {}".format(args_var))
    set_model_params(args_var,src_mods,det_mod,trans_mods,src_est,det_est,trans_est,mix_det)
    
    error = 0
    dmod_psf = det_mod.get_psf(mix_det=mix_det)
    for j in range(len(norm_rads)):
        rad = norm_rads[j]
        radm,tran,tranm = trans_mods[j].get_trans()
        pad_widths = (np.array(tran.shape)-np.array(rad.shape))//2 #assumes symmetric padding

        smod_psf = src_mods[j].get_psf()
        new_psf = combine_psfs(smod_psf,dmod_psf,are_psf=True)
        pred = convolve_psf(tran,new_psf,pad_widths,is_psf=True,warn=False)
        #error += np.mean((rad[radm]-pred[tranm])**2)
        error += np.sum((rad[radm]-pred[tranm])**2)
    error = error*0.5
    error = error/len(norm_rads)
    
#    print("Cost is {:.2e}".format(error))
    return error





import numpy as np
#import matplotlib.pyplot as plt
from skimage.measure import label,find_contours
#from PIL import Image
#from scipy.ndimage.morphology import distance_transform_edt
import csv
import sys
#from scipy.interpolate import Rbf,interp2d
from skimage.morphology import binary_opening
#from scipy import ndimage
from sklearn.linear_model import RANSACRegressor

def get_contour(rad,thresh):
    """
    Find the edge in the input radiograph.

    Parameters:
        rad (numpy.ndarray): Radiograph of a sharp edge sample
        thresh (float): The value at which a iso-valued contour (contour is the edge) is drawn

    Returns:
        numpy.ndarray: Coordinates along the longest detected contour    
    """
    contours = find_contours(rad,thresh) 
    best_contour = contours[0]
    for contour in contours:
        if(len(contour)>len(best_contour)):
            best_contour = contour
    return(best_contour)

def get_trans(rad,best_contour,trans_min,trans_max,thresh):
    """
    Compute the ideal transmission image.

    Parameters:
        rad (numpy.ndarray): Radiograph of a sharp edge sample
        best_contour (numpy.ndarray): Coordinates of the longest contour that is assumed to be the edge
        trans_min (float): Minimum transmission value
        trans_max (float): Maximum transmission value
        thresh (float): Transmission value for the edge

    Returns:
        numpy.ndarray: Ideal transmission image
    """
    window_interp = 5 #for interpolation. must be odd

    edge_thick = np.ones(rad.shape) #edge pixel will be labeled as 0
    for row,col in best_contour:
        row_floor,col_floor = int(np.floor(row)),int(np.floor(col))
        row_ceil,col_ceil = int(np.ceil(row)),int(np.ceil(col))
        edge_thick[row_floor,col_floor],edge_thick[row_floor,col_ceil] = 0,0
        edge_thick[row_ceil,col_floor],edge_thick[row_ceil,col_ceil] = 0,0

    edge_thick = binary_opening(edge_thick) #erosion followed by dilation. Rids of bright pixels in edge voxels
    rows_edge,cols_edge = np.nonzero(edge_thick==0) #Get edge pixel locations 
    
    labels,num = label(edge_thick,background=0,return_num=True)
    if(num != 2):
        raise ValueError("ERROR: Number of regions detected is {}. Two types of regions must be present in radiographs.".format(num))

    val1 = np.mean(rad[labels==1])
    val2 = np.mean(rad[labels==2])

    trans = np.zeros(rad.shape) #Sample's pixel locations will be labeled as 1
    trans[labels==0] = np.nan
    trans[labels==1] = trans_min if val1<=val2 else trans_max
    trans[labels==2] = trans_max if val1<=val2 else trans_min
      
    for row,col in best_contour:
        trans[int(round(row)),int(round(col))] = thresh

    ideal_trans = trans.copy()
    for row,col in zip(rows_edge,cols_edge):
        if(np.isnan(trans[row,col])):
            norm,ival = 0,0
            for i in range(-int((window_interp-1)/2),int((window_interp-1)/2)+1):
                for j in range(-int((window_interp-1)/2),int((window_interp-1)/2)+1):
                    row_new = row+i
                    col_new = col+j
                    if(i!=0 and j!=0 and row_new>=0 and row_new<trans.shape[0] and col_new>=0 and col_new<trans.shape[1]):
                        if(np.isnan(trans[row_new,col_new]) == False):
                            weight = 1.0/np.sqrt(i*i+j*j)
                            ival += weight*trans[row_new,col_new]
                            norm += weight
            ideal_trans[row,col] = ival/norm if norm != 0 else thresh
            if(norm == 0):
                print("WARNING: No valid value within window for interpolation") 
    return(ideal_trans) 

def get_padded_trans(ideal_trans,bdary_mask_perc,pad_factor,rad_mask):
    """
    Appropriately pad the ideal transmission image and the masks.

    Parameters:
        ideal_trans (numpy.ndarray): Ideal transmission image
        bdary_mask_perc (float): Percentage of image region that must be masked, i.e., excluded from blur estimation, close to the radiograph edges on each side (left, right, top, and bottom). Expressed as a percentage of the radiograph size.
        pad_factor (list [float,float]): Pad factor as expressed in multiples of input radiograph size
        rad_mask (numpy.ndarray): Boolean mask array over the radiograph where blur estimation is done.
    """
    bdary_mask_perc /= 100

    #Solves hw-(h-2*delta)(w-2*delta)=phw where h,w are idea_trans shape, p is bdary_mask_perc, and delta is delta_mask
    a = 4
    b = -2*(ideal_trans.shape[0]+ideal_trans.shape[1])    
    c = bdary_mask_perc*ideal_trans.shape[0]*ideal_trans.shape[1]
    delta_mask = (-b-np.sqrt(b*b-4*a*c))/(2*a)
    if delta_mask < 0:
        raise ValueError("ERROR: delta_mask is negative. This should not occur. Contact the author of this python package.")
#    print("Delta mask is ",delta_mask)

    mask = np.zeros(ideal_trans.shape).astype(bool)
    row_min = int(round(delta_mask))
    row_max = int(round(mask.shape[0]-delta_mask))
    col_min = int(round(delta_mask))
    col_max = int(round(mask.shape[1]-delta_mask))
    mask[row_min:row_max,col_min:col_max] = True
    if rad_mask is not None:
        mask = np.bitwise_and(mask,rad_mask) 
    norm_rad_mask = mask.copy()  
      
    #pad_width0 = int(ideal_trans.shape[0]*(pad_factor[0]-1)/2.0-bdary_mask_perc*mask.shape[0])
    #pad_width1 = int(ideal_trans.shape[1]*(pad_factor[1]-1)/2.0-bdary_mask_perc*mask.shape[1])
    pad_width0 = int(ideal_trans.shape[0]*(pad_factor[0]-1)/2.0)
    pad_width1 = int(ideal_trans.shape[1]*(pad_factor[1]-1)/2.0)
    colidx,rowidx = np.meshgrid(np.arange(-pad_width1,ideal_trans.shape[1]+pad_width1),np.arange(-pad_width0,ideal_trans.shape[0]+pad_width0))
    ideal_trans = np.pad(ideal_trans,((pad_width0,pad_width0),(pad_width1,pad_width1)),mode='constant',constant_values=-1)
    ideal_trans_mask = np.pad(mask,((pad_width0,pad_width0),(pad_width1,pad_width1)),mode='constant',constant_values=False)
    return(ideal_trans,ideal_trans_mask,norm_rad_mask,colidx,rowidx)

def ideal_trans_sharp_edge(norm_rad,bdary_mask_perc=5,pad_factor=[2,2],mask=None):
    """Estimate parameters of transmission model for a radiograph with a single straight sharp edge.

    This function takes as input a normalized radiograph and estimates the ideal transmission image that would have resulted if there was no blur and no noise. This function also pads the transmission image to prevent aliasing during convolution. It also produces masks for the region well within the image edges where blur model estimation is done. The transmission image and masks are packaged into a python dictionary, which is called the transmission model.  

    Parameters:
        norm_rad (numpy.ndarray): Normalized measured radiograph.
        bdary_mask_perc (float): Percentage of image region that must be masked, i.e., excluded from blur estimation, close to radiograph edges
        pad_factor (list [float,float]): Pad factor as expressed in multiples of input radiograph size
        mask (numpy.ndarray): Boolean mask array over the radiograph where blur estimation is done.
 
    Returns:
        trans_model (dict): Python dictionary containing the ideal transmission function, masks, and gradient functions.
        params_init (list): Initial parameters for ideal transmission function
        params_bounds (list of lists): Bounds on transmission function parameters.  
    """

    #pad_factor: pad to multiple of the input image size

    trans_min = max(0.0,np.min(norm_rad))
    trans_max = min(1.0,np.max(norm_rad))
    renorm_rad = (norm_rad-trans_min)/(trans_max-trans_min)
    best_contour = get_contour(renorm_rad,0.5)

    ideal_trans = get_trans(renorm_rad,best_contour,0.0,1.0,0.5)

    x,y = [],[]
    for row,col in best_contour:
        x.append(col)
        y.append(row)
    coeff = np.polyfit(x,y,1)
    line1 = np.poly1d(coeff)   

    ideal_trans,ideal_trans_mask,norm_rad_mask,colidx,rowidx = get_padded_trans(ideal_trans,bdary_mask_perc,pad_factor,mask)    
 
    linediff1 = rowidx - line1(colidx)
            
    for reg in [linediff1>0,linediff1<=0]:
        val = np.mean(ideal_trans[np.bitwise_and(ideal_trans>=0,reg)])
        ideal_trans[np.bitwise_and(ideal_trans==-1,reg)] = 0.0 if val<0.5 else 1.0
        
    if(np.any(np.isnan(ideal_trans))):
        raise ValueError("ERROR: Nan detected in the ideal radiograph image")

    X_rows = ideal_trans[ideal_trans_mask].size
    X0 = 1-ideal_trans[ideal_trans_mask].reshape(X_rows,1)
    X1 = ideal_trans[ideal_trans_mask].reshape(X_rows,1)
    X = np.hstack((X0,X1))
    y = norm_rad[norm_rad_mask].reshape(X_rows,1)
    reg = RANSACRegressor(min_samples=10,residual_threshold=0.1)
    reg.fit(X,y)
    params_init = [float(reg.predict(np.array([[1,0]]))),float(reg.predict(np.array([[0,1]])))]    
    
    z = ideal_trans[ideal_trans_mask]
    z_sq = ideal_trans[ideal_trans_mask]**2
    A = np.zeros((2,2))
    A[0,0] = np.sum(1-2*z+z_sq)
    A[0,1] = np.sum(z-z_sq)
    A[1,0] = A[0,1]
    A[1,1] = np.sum(z_sq)
    b = np.zeros(2)
    b[0] = np.sum(norm_rad[norm_rad_mask]*(1-z))
    b[1] = np.sum(norm_rad[norm_rad_mask]*z)
    #print("old trans init params {}".format(np.matmul(np.linalg.inv(A),b).tolist()))
    params_bounds = [[-1.0,0.5],[0.5,2.0]]

    #print("trans init params {}".format(params_init))

    trans_model = {}
    trans_model['norm_rad_mask'] = norm_rad_mask 
    trans_model['ideal_trans'] = lambda params: params[0]+(params[1]-params[0])*ideal_trans
    trans_model['ideal_trans_mask'] = ideal_trans_mask
    trans_model['ideal_trans_grad'] = lambda params: [1.0-ideal_trans,ideal_trans]

    return trans_model,params_init,params_bounds 

def ideal_trans_perp_corner(norm_rad,bdary_mask_perc=5,pad_factor=[2,2],mask=None,perp_mask_perc=5):
    """Estimate parameters of transmission model for a radiograph with a sharp perpendicular corner.

    This function takes as input a normalized radiograph and estimates the ideal transmission image that would have resulted if there was no blur and no noise. This function also pads the transmission image to prevent aliasing during convolution. It also produces masks for the region well within the image edges where blur model estimation is done. The transmission image and masks are packaged into a python dictionary, which is called the transmission model.  

    Parameters:
        norm_rad (numpy.ndarray): Normalized measured radiograph.
        bdary_mask_perc (float): Percentage of image region that must be masked, i.e., excluded from blur estimation, close to radiograph edges
        pad_factor (list [float,float]): Pad factor as expressed in multiples of input radiograph size
        mask (numpy.ndarray): Boolean mask array over the radiograph where blur estimation is done.
 
    Returns:
        trans_model (dict): Python dictionary containing the ideal transmission function, masks, and gradient functions.
        params_init (list): Initial parameters for ideal transmission function
        params_bounds (list of lists): Bounds on transmission function parameters.  
    """

    trans_min = max(0.0,np.min(norm_rad))
    trans_max = min(1.0,np.max(norm_rad))
    renorm_rad = (norm_rad-trans_min)/(trans_max-trans_min)
    best_contour = get_contour(renorm_rad,0.5)
    
    ideal_trans = get_trans(renorm_rad,best_contour,0.0,1.0,0.5)

    conlen = len(best_contour)
    x,y = [],[]
    for row,col in best_contour[:conlen//4]:
        x.append(col)
        y.append(row)
    line1_coeff = np.polyfit(x,y,1)
    line1 = np.poly1d(line1_coeff)   

    x,y = [],[]
    for row,col in best_contour[-(conlen//4):]:
        x.append(col)
        y.append(row)
    line2_coeff = np.polyfit(x,y,1)
    line2 = np.poly1d(line2_coeff)   
    
    ideal_trans,ideal_trans_mask,norm_rad_mask,colidx,rowidx = get_padded_trans(ideal_trans,bdary_mask_perc,pad_factor,mask)    
        
    linediff1 = rowidx - line1(colidx)
    linediff2 = rowidx - line2(colidx)
   
    region1 = np.bitwise_and(linediff1<=0,linediff2<=0)
    region2 = np.bitwise_and(linediff1>0,linediff2<=0)
    region3 = np.bitwise_and(linediff1<=0,linediff2>0)
    region4 = np.bitwise_and(linediff1>0,linediff2>0)
            
    for reg in [region1,region2,region3,region4]:
        val = np.mean(ideal_trans[np.bitwise_and(ideal_trans>=0.0,reg)])
        ideal_trans[np.bitwise_and(ideal_trans==-1,reg)] = 0.0 if val<0.5 else 1.0
  
    perp_mask_perc /= (2*100)
    col_inter = (line1_coeff[1]-line2_coeff[1])/(line2_coeff[0]-line1_coeff[0])
    row_inter = line1_coeff[0]*col_inter+line1_coeff[1]
    circmask = (colidx-col_inter)**2+(rowidx-row_inter)**2<=(perp_mask_perc*max(norm_rad.shape))**2
    ideal_trans_mask[circmask] = False   
    colidx,rowidx = np.meshgrid(np.arange(0,norm_rad.shape[1]),np.arange(0,norm_rad.shape[0]))
    circmask = (colidx-col_inter)**2+(rowidx-row_inter)**2<=(perp_mask_perc*max(norm_rad.shape))**2
    norm_rad_mask[circmask] = False

    if(np.any(np.isnan(ideal_trans))):
        raise ValueError("ERROR: Nan detected in the ideal radiograph image")

    X_rows = ideal_trans[ideal_trans_mask].size
    X0 = 1-ideal_trans[ideal_trans_mask].reshape(X_rows,1)
    X1 = ideal_trans[ideal_trans_mask].reshape(X_rows,1)
    X = np.hstack((X0,X1))
    y = norm_rad[norm_rad_mask].reshape(X_rows,1)
    reg = RANSACRegressor(min_samples=10,residual_threshold=0.1)
    reg.fit(X,y)
    params_init = [float(reg.predict(np.array([[1,0]]))),float(reg.predict(np.array([[0,1]])))]    
 
    z = ideal_trans[ideal_trans_mask]
    z_sq = ideal_trans[ideal_trans_mask]**2
    A = np.zeros((2,2))
    A[0,0] = np.sum(1-2*z+z_sq)
    A[0,1] = np.sum(z-z_sq)
    A[1,0] = A[0,1]
    A[1,1] = np.sum(z_sq)
    b = np.zeros(2)
    b[0] = np.sum(norm_rad[norm_rad_mask]*(1-z))
    b[1] = np.sum(norm_rad[norm_rad_mask]*z)
#    print("old trans init params {}".format(np.matmul(np.linalg.inv(A),b).tolist()))
#    print("trans init params {}".format(params_init))
    
    trans_model = {}
    trans_model['norm_rad_mask'] = norm_rad_mask 
    trans_model['ideal_trans'] = lambda params: params[0]+(params[1]-params[0])*ideal_trans
    trans_model['ideal_trans_mask'] = ideal_trans_mask
    trans_model['ideal_trans_grad'] = lambda params: [1.0-ideal_trans,ideal_trans]
    params_init = np.matmul(np.linalg.inv(A),b).tolist()
    params_bounds = [[-1.0,0.5],[0.5,2.0]]

    return trans_model,params_init,params_bounds




import setuptools

with open('VERSION','r') as fh:
    version = fh.read().strip()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='pysaber',
      version=version,
      description='Python package that implements a systems approach to blur estimation and reduction (SABER)',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/sabersw',
      author='K. Aditya Mohan',
      author_email='mohan3@llnl.gov',
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=['numpy','pyyaml','scipy','scikit-image','scikit-learn','matplotlib'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      )



from pysaber.mainsr import estimate_blur,get_source_psf,get_detector_psf,get_effective_psf,apply_blur_psfs,get_trans_fit,get_trans_masks
from pysaber.deblursr import wiener_deblur,least_squares_deblur

